from pygears import alternative, gear
from pygears.typing import Queue


@gear
async def alternate_queues(din0: Queue, din1: Queue) -> b'(din0, din1)':
    """Alternates the propagation of input transactions to their output pairs"""
    async for d in din0:
        yield d, None
    async for d in din1:
        yield None, d


@alternative(alternate_queues)
@gear
async def alternate_queues_multi(*din: Queue) -> b'(din[0], ) * len(din)':
    for i, d in enumerate(din):
        out_res = [None] * len(din)
        async for data in d:
            out_res[i] = data
            yield out_res



from ..ir_utils import HDLVisitor, add_to_list, ir, res_false, IrVisitor, IrExprVisitor


class FuncCallExprFinder(IrExprVisitor):
    def __init__(self, called_funcs):
        self.called_funcs = called_funcs

    def visit_FunctionCall(self, node: ir.FunctionCall):
        self.called_funcs.add(node.name)
        super().visit_FunctionCall(node)


class FuncCallFinder(IrVisitor):
    def __init__(self):
        super().__init__()
        self.called_funcs = set()
        self.expr_visit = FuncCallExprFinder(self.called_funcs)

    def Expr(self, expr):
        self.expr_visit.visit(expr)


class RemoveDeadCode(HDLVisitor):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.called_funcs = set()

    def AssignValue(self, node):
        return node

    def FuncReturn(self, node):
        return node

    def FunctionCall(self, node: ir.FunctionCall):
        self.called_funcs.add(node.name)
        return node

    def BaseBlock(self, block: ir.BaseBlock):
        stmts = []
        for stmt in block.stmts:
            add_to_list(stmts, self.visit(stmt))

        block.stmts = stmts
        return block

    # def ExprStatement(self, stmt: ir.ExprStatement):
    #     return None

    # TODO: Implement things properly for IfElseBlock (What if middle elif is missing?)

    def HDLBlock(self, node):
        branches = []
        for b in node.branches:
            if b.test == res_false:
                continue

            if not b.stmts:
                continue

            branches.append(self.visit(b))

        node.branches = branches

        return node


def remove_dead_code(modblock, ctx):
    return RemoveDeadCode(ctx).visit(modblock)


def find_called_funcs(modblock, ctx):
    v = FuncCallFinder()
    v.visit(modblock)
    return v.called_funcs



from pygears import alternative, gear
from pygears.typing import Queue
from pygears.lib.quenvelope import quenvelope
from pygears.lib.czip import czip
from pygears.lib.project import project
from pygears.lib.fmap import fmap as common_fmap


@alternative(common_fmap)
@gear(enablement=b'issubclass(din, Queue)')
def queuemap(din, *, f, lvl=1, fcat=czip, balance=None, common_balance=True):
    queue_lvl = din.dtype.lvl
    fmap_lvl = min(lvl, queue_lvl)
    lvl -= fmap_lvl

    env = din | quenvelope(lvl=fmap_lvl)
    data = din | project(lvl=fmap_lvl)

    if lvl > 0:
        f = common_fmap(f=f, lvl=lvl, fcat=fcat, balance=balance)

    dout = f(data)

    if not isinstance(dout, tuple):
        dout = (dout, )

    for i, d in enumerate(dout):
        if d is None:
            raise TypeError(
                f'Module "{f.__name__}" passed to queuemap returned nothing on its ouput {i}'
            )

    if balance is not None and common_balance:
        env = env | balance

    cat_dout = []
    for d in dout:
        benv = env
        if balance is not None and not common_balance:
            benv = env | balance

        cat_dout.append(fcat(benv, d))

    # Cast to remove tuple Unit from resulting cart tuple
    ret = [d >> Queue[d.dtype[0][1], d.dtype.lvl] for d in cat_dout]

    return tuple(ret)



from ..ir_utils import ir
from ..cfg import CfgDfs
from ..cfg_util import remove_node
from ..ir_utils import ir, IrRewriter, add_to_list, IrVisitor


class ResolveRegInits(IrRewriter):
    def __init__(self, ctx, scope_map=None):
        super().__init__()
        self.ctx = ctx

    def AssignValue(self, node: ir.AssignValue):
        if not isinstance(node.target, ir.Name):
            return node

        name = node.target.name

        if name not in self.ctx.reaching[id(node)]['in']:
            obj = self.ctx.scope[name]
            # If this is a register variable and assigned value is a literal value (ResExpr)
            if (isinstance(obj, ir.Variable) and obj.reg and isinstance(node.val, ir.ResExpr)
                    and node.val.val is not None):
                init_val = ir.CastExpr(node.val, obj.dtype)
                if obj.val is None:
                    node = ir.RegReset(obj)
                    obj.val = init_val
                    obj.any_init = False
                elif obj.val != init_val and obj.any_init:
                    breakpoint()
                    print('Hier?')

        return node


# class ResolveRegInits(CfgDfs):
#     def __init__(self, ctx):
#         self.ctx = ctx
#         super().__init__()

#     # TODO: How to detect registers used uninitialized?
#     def enter_AssignValue(self, node):
#         irnode: ir.AssignValue = node.value
#         if not isinstance(irnode.target, ir.Name):
#             return

#         breakpoint()
#         name = irnode.target.name

#         if name not in self.ctx.reaching[id(node.value)]['in']:
#             obj = self.ctx.scope[name]
#             # If this is a register variable and assigned value is a literal value (ResExpr)
#             if (isinstance(obj, ir.Variable) and obj.reg and isinstance(irnode.val, ir.ResExpr)
#                     and irnode.val.val is not None):
#                 init_val = ir.CastExpr(irnode.val, obj.dtype)
#                 if obj.val is None or obj.val == init_val:
#                     node.value = ir.RegReset(obj)
#                     obj.val = init_val
#                     obj.any_init = False
#                 elif obj.any_init:
#                     breakpoint()
#                     print('Hier?')


def infer_registers(modblock, ctx):
    return ResolveRegInits(ctx).visit(modblock)



import time
import atexit
import tempfile
from math import ceil
import jinja2
import array
import math
import os
import socket
import logging
from pygears.definitions import ROOT_DIR
from subprocess import DEVNULL, Popen
from pygears import reg
from pygears.sim.extens.sim_extend import SimExtend
from pygears.sim import SimPlugin
from pygears.util.fileio import save_file
from pygears.sim import log

from pygears.conf import inject, Inject

CMD_SYS_RESET = 0x80000000
CMD_SET_DATA = 0x40000000
CMD_RESET = 0x20000000
CMD_FORWARD = 0x10000000
CMD_CYCLE = 0x08000000
CMD_READ = 0x04000000
CMD_ACK = 0x02000000
CMD_FINISH = 0x01000000


class CosimulatorStartError(Exception):
    pass


class CosimulatorUnavailable(Exception):
    pass


def register_exten():
    if SVSock not in reg['sim/extens']:
        reg['sim/extens'].append(SVSock)
        if reg['sim/simulator'] is not None:
            SVSock(top=None)


@inject
def register_intf(desc, intfs=Inject('sim/svsock/intfs')):
    register_exten()
    intfs.append(desc)
    return len(intfs) - 1


def u32_repr_gen(data, dtype):
    for i in range(ceil(dtype.width / 32)):
        yield data & 0xffffffff
        data >>= 32


def u32_repr(data, dtype):
    return array.array('I', u32_repr_gen(dtype(data).code(), dtype))


def u32_bytes_to_int(data):
    arr = array.array('I')
    arr.frombytes(data)
    val = 0
    for val32 in reversed(arr):
        val <<= 32
        val |= val32

    return val


def u32_bytes_decode(data, dtype):
    return dtype.decode(u32_bytes_to_int(data) & ((1 << dtype.width) - 1))


class SVSock(SimExtend):
    @inject
    def __init__(self, run=Inject('sim/svsock/run'), **kwds):
        reg['sim/svsock/server'] = self
        self.run_cosim = run
        self.kwds = kwds
        self.sock = None
        self.conn = None
        self.cosim_pid = None
        super().__init__()
        atexit.register(self.finish)

    @property
    @inject
    def outdir(self, outdir=Inject('results-dir')):
        return os.path.join(outdir, 'svsock')

    @inject
    def build(self, intfs=Inject('sim/svsock/intfs')):

        base_addr = os.path.dirname(__file__)
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(base_addr),
                                 trim_blocks=True,
                                 lstrip_blocks=True)
        env.globals.update(zip=zip,
                           int=int,
                           print=print,
                           issubclass=issubclass)

        context = {'port': self.port}
        for phase in [
                'declaration', 'init', 'set_data', 'read', 'ack', 'reset',
                'sys_reset'
        ]:
            context[phase] = {}
            for i, intf in enumerate(intfs):
                if not hasattr(intf, phase):
                    continue

                phase_sv = getattr(intf, phase)()

                if phase_sv is not None:
                    context[phase][i] = phase_sv

        res = env.get_template('svsock_top.j2').render(context)
        save_file('_top.sv', self.outdir, res)

    def sendall(self, pkt):
        self.conn.sendall(pkt)

    def send_cmd(self, req):
        pkt = req.to_bytes(4, byteorder='little')
        self.conn.sendall(pkt)

    def recv(self, size):
        return self.conn.recv(size)

    def dtype_send(self, data, dtype):
        pkt = u32_repr(data, dtype).tobytes()
        self.sendall(pkt)

    def dtype_recv(self, dtype):
        buff_size = math.ceil(dtype.width / 8)
        if buff_size < 4:
            buff_size = 4
        if buff_size % 4:
            buff_size += 4 - (buff_size % 4)

        data = self.recv(buff_size)
        return u32_bytes_decode(data, dtype)

    @inject
    def invoke_cosim(self, intfs=Inject('sim/svsock/intfs')):
        dpi_path = os.path.abspath(os.path.join(ROOT_DIR, 'sim', 'dpi'))

        context = {}
        context['files'] = [os.path.join(dpi_path, 'sock.sv')]
        context['includes'] = []

        for param in ['files', 'includes']:
            for i, intf in enumerate(intfs):
                if not hasattr(intf, param):
                    continue

                param_val = getattr(intf, param)()

                if param_val is not None:
                    context[param].extend(param_val)

        context['includes'].extend([dpi_path, self.outdir])
        context['files'].extend([os.path.join(self.outdir, '_top.sv')])

        if not reg['sim/svsock/backend']:
            raise CosimulatorStartError('No registered cosimulators')

        cosim_pid = None
        for b in reg['sim/svsock/backend'].values():
            try:
                cosim_pid = b(outdir=self.outdir,
                              files=context['files'],
                              includes=context['includes'],
                              makefile=not self.run_cosim)
            except CosimulatorUnavailable:
                pass
            else:
                break
        else:
            raise CosimulatorStartError(
                f'No available cosimulator executables found for any of the plugins: '
                f'{",".join(reg["sim/svsock/backend"].keys())}')

        return cosim_pid

    def before_run(self, sim):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        if self.run_cosim:
            import uuid
            filename = str(uuid.uuid4())
        else:
            filename = "svsock.s"

        self.port = os.path.join(tempfile.gettempdir(), filename)

        if os.path.exists(self.port):
            os.remove(self.port)
        self.sock.bind(self.port)

        # Listen for incoming connections
        # self.sock.listen(len(self.gear.in_ports) + len(self.gear.out_ports))
        self.sock.listen(1)

        self.build()

        # if self.rebuild:
        #     self.build()
        # else:
        #     self.kwds['nobuild'] = True

        if not self.run_cosim:
            self.invoke_cosim()
            self.conn, addr = self.sock.accept()
        else:
            self.sock.settimeout(5)

            self.cosim_pid = self.invoke_cosim()

            ret = None
            while ret is None:
                try:
                    self.conn, addr = self.sock.accept()
                    break
                except socket.timeout:
                    ret = self.cosim_pid.poll()
                    if ret is not None:
                        log.error(
                            f'Cosimulator error: {ret}. Check log File "{self.outdir}/log.log"'
                        )
                        raise CosimulatorStartError

        msg = self.conn.recv(1024)
        port_name = msg.decode()

        log.debug(f"Connection received for {port_name}")

    def finish(self):
        if self.sock:
            if self.conn:
                try:
                    self.send_cmd(CMD_FINISH)
                    time.sleep(0.5)
                except BrokenPipeError:
                    pass

            log.info(f'Done. Closing the socket...')
            self.sock.close()
            time.sleep(1)

            if self.cosim_pid is not None:
                self.cosim_pid.terminate()

            self.sock = None
            self.cosim_pid = None
            self.conn = None
            atexit.unregister(self.finish)

    def after_cleanup(self, sim):
        self.finish()


class SVSockPlugin(SimPlugin):
    @classmethod
    def bind(cls):
        reg.confdef('sim/svsock/backend', default={})
        reg.confdef('sim/svsock/run', default=True)
        reg['sim/svsock/intfs'] = []
        reg['sim/svsock/server'] = None



from pygears import alternative, gear
from pygears.sim import clk
from pygears.typing import Queue, Union


@gear(enablement=b'not all(typeof(d, Queue) for d in din)')
async def priority_mux(*din) -> b'Union[din]':
    """Takes in a tuple of interfaces and passes any active one to the output. If
    two or more inputs are given at the same time, the input having the highest
    priority (higher in the list of inputs) will take precedence and will be
    passed to the output.

    Returns:
        A :class:`Union` type where the ``ctrl`` field signalizes which input was
          passed.
    """
    for i, d in enumerate(din):
        if not d.empty():
            async with d as item:
                yield (item, i)
            break
    else:
        await clk()


def prio_mux_queue_type(dtypes):
    utypes = (dtypes[0][0], ) * len(dtypes)
    return Queue[Union[utypes], dtypes[0].lvl]


@alternative(priority_mux)
@gear(enablement=b'all(typeof(d, Queue) for d in din)',
      hdl={'impl': 'priority_mux'})
async def priority_mux_queue(*din) -> b'prio_mux_queue_type(din)':
    """Priority mux alternative which operates on queues"""
    for i, d in enumerate(din):
        if not d.empty():
            async for (data, eot) in d:
                yield ((data, i), eot)
            break
    else:
        await clk()



from pygears import gear, IntfEmpty
from pygears.typing import Bool
from pygears.sim import clk


@gear
async def dreg(din, *, init=None) -> b'din':
    data = din.dtype() if init is None else din.dtype(init)

    valid = Bool(False) if init is None else Bool(True)

    while True:
        if valid:
            yield data

            try:
                data = din.get_nb()
                valid = True
            except IntfEmpty:
                valid = False
        else:
            data = await din.get()
            valid = True
            await clk()


@gear
def regout(din, *, f):
    return din | f | dreg



from pygears import gear
from pygears.typing import Queue, Tuple, Uint, code, Array
from .ccat import ccat
from .czip import czip


@gear
def queue_wrap_from(din, qdin, *, fcat=czip):
    cat_data = fcat(qdin, din)

    return ccat(cat_data['data'][1], cat_data['eot']) \
        | Queue[din.dtype, qdin.dtype.lvl]


@gear
async def sot_queue(din: Queue['data', 'lvl'], *,
                    lvl=b'lvl') -> Tuple[b'din.data', b'din.eot']:
    sot = din.dtype.eot.max

    async for (data, eot) in din:
        yield (data, sot)

        neot = (~eot) << 1

        sot_arr = Array[Uint[1], lvl](None)
        for i in range(lvl):
            sot_arr[i] = eot[i] if i == 0 else sot[i] & neot[i]

        sot = code(sot_arr, Uint)



from pygears import alternative, gear
from pygears.lib import ccat
from pygears.typing import Integer, Tuple


@gear
async def max2(
        din: Tuple[Integer['N1'], Integer['N2']]) -> b'max(din[0], din[1])':
    """Finds the highest of the two values"""
    async with din as data:
        yield max(data)


@alternative(max2)
@gear
def max22(din0: Integer, din1: Integer):
    return ccat(din0, din1) | max2



import fnmatch
import functools
from string import Template

from pygears import PluginBase, reg
from pygears.conf import Inject, inject
from pygears.core.port import OutPort, InPort
# from .util import svgen_typedef

dti_spy_connect_t = Template("""
dti_spy #(${intf_name}_t) _${intf_name}(clk, rst);
assign _${intf_name}.data = ${conn_name}.data;
assign _${intf_name}.valid = ${conn_name}.valid;
assign _${intf_name}.ready = ${conn_name}.ready;""")


class VIntfGen:
    def __init__(self, intf):
        self.intf = intf

    @property
    def basename(self):
        return self.intf.basename

    @property
    def outname(self):
        return self.intf.outname

    @property
    @functools.lru_cache()
    def traced(self):
        return any(
            fnmatch.fnmatch(self.intf.name, p)
            for p in reg['debug/trace'])

    @inject
    def get_inst(self,
                 template_env,
                 spy_connect_t=Inject('svgen/spy_connection_template')):

        if self.intf.producer is None:
            return

        inst = []
        if not self.intf.is_port_intf:
            inst.append(self.get_intf_def(self.basename, 1, template_env))

        if self.intf.is_broadcast:
            inst.extend([
                self.get_intf_def(self.outname, len(self.intf.consumers),
                                  template_env),
                self.get_bc_module(template_env)
            ])

        for i, cons_port in enumerate(self.intf.consumers):
            if isinstance(cons_port, OutPort):
                if self.intf.is_broadcast or isinstance(self.intf.producer, InPort):
                    inst.append(self.get_connect_module(template_env, index=i))

            # if self.traced:
            #     for i in range(len(self.intf.consumers)):
            #         intf_name = f'{self.outname}_{i}'
            #         conn_name = f'{self.outname}[{i}]'
            #         inst.extend(
            #             svgen_typedef(self.intf.dtype, intf_name).split('\n'))

            #         inst.extend(
            #             spy_connect_t.substitute(
            #                 intf_name=intf_name,
            #                 conn_name=conn_name).split('\n'))

        # if self.traced:
        #     inst.extend(
        #         svgen_typedef(self.intf.dtype, self.basename).split('\n'))

        #     inst.extend(
        #         spy_connect_t.substitute(intf_name=self.basename,
        #                                  conn_name=self.basename).split('\n'))

        return '\n'.join(inst)

    def get_intf_def(self, name, size, template_env):
        ctx = {
            'name': name,
            'width': self.intf.dtype.width,
            'size': size,
            'type': str(self.intf.dtype)
        }

        return template_env.snippets.intf_inst(**ctx)

    def get_connect_module(self, template_env, index=None):
        if self.intf.is_broadcast:
            port_name = self.intf.consumers[index].basename
        else:
            port_name = self.intf.consumers[0].basename

        connect_context = {
            'intf_name': self.outname,
            'width': self.intf.dtype.width,
            'index': index if self.intf.is_broadcast else None,
            'port_name': port_name
        }

        return template_env.snippets.bc_to_out_port_connect(**connect_context)

    def get_bc_module(self, template_env):
        inst_name = f'bc_{self.basename}'
        if inst_name.endswith('_if_s'):
            inst_name = inst_name[:-len('_if_s')]

        bc_context = {
            'rst_name': 'rst',
            'module_name': 'bc',
            'inst_name': inst_name,
            'param_map': {
                'SIZE': len(self.intf.consumers),
                'WIDTH': self.intf.dtype.width
            },
            'port_map': {
                'din': (self.basename, None, None),
                'dout': (self.outname, None, None)
            }
        }

        return template_env.snippets.module_inst(**bc_context)


class VGenIntfPlugin(PluginBase):
    @classmethod
    def bind(cls):
        pass
        # reg['svgen/spy_connection_template'] = dti_spy_connect_t



import inspect
import types
import textwrap
import ast
import os


def get_function_source(func):
    try:
        source = inspect.getsource(func)
    except (OSError, TypeError):
        try:
            source = func.__source__
        except AttributeError:
            return ''

    return textwrap.dedent(source)


def get_short_lambda_ast(lambda_func):
    """Return the source of a (short) lambda function.
    If it's impossible to obtain, returns None.

    taken from: http://xion.io/post/code/python-get-lambda-code.html
    """
    try:
        source_lines, _ = inspect.getsourcelines(lambda_func)
    except (IOError, TypeError):
        return None

    # skip `def`-ed functions and long lambdas
    if len(source_lines) != 1:
        return None

    source_text = os.linesep.join(source_lines).strip()

    # find the AST node of a lambda definition
    # so we can locate it in the source code
    source_ast = ast.parse(source_text)
    lambda_node = next((node for node in ast.walk(source_ast) if isinstance(node, ast.Lambda)),
                       None)
    if lambda_node is None:  # could be a single line `def fn(x): ...`
        return None

    return lambda_node


def is_lambda_function(obj):
    return isinstance(obj, types.LambdaType) and obj.__name__ == (lambda: None).__name__


def get_function_ast(func):
    if is_lambda_function(func):
        lambda_ast = get_short_lambda_ast(func)
        lambda_ast.body = [ast.Return(lambda_ast.body)]
        return ast.fix_missing_locations(lambda_ast)
    else:
        src = get_function_source(func)
        if not src:
            raise SyntaxError(f"Function '{func.__qualname__}()' not supported")

        return ast.parse(src).body[0]

def is_func_empty(func):
    # TODO: We are doing AST generation at least twice, performance impact?
    body = get_function_ast(func).body
    if isinstance(body[0], ast.Pass):
        return True

    if isinstance(body[0], ast.Expr) and isinstance(body[0].value, ast.Ellipsis):
        return True

    return False


def get_property_type(prop):
    assert isinstance(prop, property)

    return prop.fget.__annotations__['return']


def add_to_list(orig_list, extension):
    if extension:
        if not isinstance(extension, list):
            orig_list.append(extension)
        else:
            orig_list.extend(extension)



import os
from .hdlfile import HDLFileResolver
from pygears.conf import inject, Inject

class BlackBoxResolver(HDLFileResolver):
    @inject
    def __init__(self, node):
        self.node = node

    @property
    def impl_parse(self):
        return [os.path.splitext(self.file_basename)[0], [], [], {}]

    @property
    def files(self):
        return []



import ast
from . import Context, FuncContext, ir, node_visitor, visit_ast, visit_block
from ..cfg_ast import reaching


@node_visitor(ast.AsyncFunctionDef)
def parse_async_func(node, ctx: Context):

    body_ast, cfg, r, registers = reaching(node)

    ctx.reaching = r
    ctx.registers = registers

    ctx.reset_states = {}
    for reg in registers:
        ctx.reset_states[reg] = set()

    return visit_block(ir.Module(), node.body, ctx)


@node_visitor((ast.Lambda, ast.FunctionDef))
def _(node, ctx: FuncContext):
    if not isinstance(ctx, FuncContext):
        raise Exception('Unsupported')

    block = ir.FuncBlock(stmts=[],
                         args=ctx.signature,
                         name=ctx.funcref.name,
                         ret_dtype=ctx.ret_dtype)

    body_ast, cfg, r, registers = reaching(node)
    ctx.reaching = r
    ctx.registers = registers

    ret = visit_block(block, node.body, ctx)

    # TODO: Why is this necessary?
    block.ret_dtype = ctx.ret_dtype

    return ret



import inspect
from pygears import reg, Intf
from pygears.core.gear_decorator import create_gear_definition
from pygears.core.gear_decorator import FunctionMaker
from pygears.core.util import doublewrap, get_function_context_dict
from pygears.core.util import is_standard_func
from pygears.util.utils import gather


def is_datagear(func):
    return hasattr(func, 'exec')


def get_datagear_func(func):
    return func.exec


# TODO: "outnames" not working
@doublewrap
def datagear(func, **meta_kwds):
    if not is_standard_func(func):
        raise Exception('Only regular functions can be converted to a @datagear.')

    paramspec = inspect.getfullargspec(func)

    invocation = ['*_data']

    for name in paramspec.kwonlyargs:
        invocation.append(f'{name}={name}')

    body = f'''async with gather({",".join(paramspec.args)}) as _data:
        yield datafunc({",".join(invocation)})'''

    # TODO: Small optimization point to avoid async context manager. This needs to be supported by the compiler
    # invocation = [f'await {a}.pull()' for a in paramspec.args]
    # for name in paramspec.kwonlyargs:
    #     invocation.append(f'{name}={name}')

    # ack = ','.join(f'{a}.ack()' for a in paramspec.args)

    # body = f'''
    # yield datafunc({",".join(invocation)})
    # {ack}
    # '''

    execdict = {'datafunc': func, 'gather': gather}
    execdict.update(get_function_context_dict(func))

    gear_func = FunctionMaker.create(obj=func,
                                     body=body,
                                     evaldict=execdict,
                                     isasync=True,
                                     addsource=True)
    gear_func.exec = func

    return create_gear_definition(gear_func)



from pygears import gear


@gear
async def priority_mux_valve(*din) -> b'Union[din]':
    """Similar to the ``priority_mux`` gear, but the data is consumed for all
    input interfaces when any data is sent to the output i.e. the output `ready`
    is propagated to all input `ready` signals.
    """



from .base import EnumerableGenericMeta, typeof, is_type, TemplateArgumentsError
from .base import class_and_instance_method

# TODO: Check why array is specified when no length is specified


# TODO: Should we define addition?
class ArrayType(EnumerableGenericMeta):
    def __new__(cls, name, bases, namespace, args=None):
        cls = super().__new__(cls, name, bases, namespace, args)

        if not cls.specified:
            return cls

        if not isinstance(cls.args[1], int):
            err = None
            try:
                cls.__args__[1] = int(cls.args[1])
            except TypeError as e:
                err = e

            if err:
                raise TemplateArgumentsError(
                    f'Second argument to the "Array" type must be integer, not "{repr(cls.args[1])}'
                )

        return cls

    def keys(self):
        """Returns a list of keys that can be used for indexing :class:`Array` [T, N] type. Number of keys equals to the number of elements N.

        >>> Array[Uint[2], 5].keys()
        [0, 1, 2, 3, 4]
        """

        return list(range(self.args[1]))

    @property
    def width(self):
        return self._args[0].width * int(self._args[1])

    def __len__(self):
        if isinstance(self._args[1], int):
            return self._args[1]
        elif not self.specified:
            raise TypeError(f"Length of type '{self}' is generic (unspecifed)")
        else:
            raise Exception("Unexpected error")

    # TODO: Remove this
    @property
    def dtype(self):
        return self._args[0]

    @property
    def data(self):
        return self._args[0]

    def __getitem__(self, index):
        """If a single element is supplied for index, returns type T. If a slice is suplied for index, an :class:`Array` type is returned with a number of elements equal to the slice size.

        >>> Array[Uint[2], 5][3]
        Uint[2]

        >>> Array[Uint[2], 5][2:4]
        Array[Uint[2], 2]
        """

        if not self.specified:
            return super().__getitem__(index)

        if isinstance(index, int):
            if index >= len(self):
                raise IndexError

            return self._args[0]

        index = self.index_norm(index)

        if len(index) == 1 and not isinstance(index[0], slice):
            if index[0] >= len(self):
                raise IndexError

            return self.args[0]
        else:
            width = 0
            for i in index:
                if isinstance(i, slice):
                    if (i.stop == 0) or (i.stop - i.start > len(self)):
                        raise IndexError
                    width += i.stop - i.start
                else:
                    if i >= len(self):
                        raise IndexError
                    width += 1

            return Array[self.args[0], width]

    def __str__(self):
        if self.args:
            return f'Array[{str(self.args[0])}, {self.args[1]}]'
        else:
            return super().__str__()


class Array(list, metaclass=ArrayType):
    """Generic container datatype that holds N instances of type T

    Generic parameters:
       T: Type of the :class:`Array` [T, N] elements

       N: Number of elements in the :class:`Array` [T, N]

    Concrete data type is obtained by indexing::

        u16_4 = Array[Uint[16], 4]

    """
    __parameters__ = ['T', 'N']

    def __init__(self, val: tuple = None):
        t = type(self).data

        if val is None:
            array_tpl = (t(), ) * len(type(self))
        else:
            array_tpl = (v if typeof(type(v), t) or v is None else t(v) for v in val)

        return super().__init__(array_tpl)

    def __eq__(self, other):
        t_other = type(other)
        if not is_type(t_other):
            return super().__eq__(other)

        return type(self) == t_other and super().__eq__(other)

    def __ne__(self, other):
        t_other = type(other)
        if not is_type(t_other):
            return super().__ne__(other)

        return type(self) != t_other or super().__ne__(other)

    @class_and_instance_method
    def __getitem__(self, key):
        if isinstance(key, int):
            return super().__getitem__(key)
        elif isinstance(key, str):
            try:
                return super().__getitem__(type(self).fields.index(key))
            except ValueError:
                raise TypeError(f'Tuple "{repr(self)}" has no field "{key}"')

        key_norm = type(self).index_norm(key)

        if len(key_norm) == 1:
            if isinstance(key_norm[0], slice):
                tout = type(self)[key_norm]
                return tout(super().__getitem__(key_norm[0]))
            else:
                return super(Array, self).__getitem__(key_norm[0])
        else:
            tout = type(self)[key_norm]
            elems = []
            for i in key_norm:
                elems.extend(super().__getitem__(i))

            return tout(elems)

    @class_and_instance_method
    def subs(self, path, val):
        if isinstance(path, tuple):
            if len(path) > 1:
                val = self[path[0]].subs(path[1:], val)

            path = path[0]

        return type(self)([self[i] if i != path else val for i in range(len(self))])

    def __hash__(self):
        return hash(tuple(self))

    def code(self):
        w_dtype = type(self).data.width
        ret = 0
        for d in reversed(self):
            ret <<= w_dtype
            if d is not None:
                ret |= d.code()

        return ret

    @property
    def unknown(self):
        return any(v is None or getattr(v, 'unknown', False) for v in self)

    @classmethod
    def decode(cls, val):
        ret = []
        val = int(val)
        mask = int(cls.data.width * '1', 2)
        for t in cls:
            ret.append(t.decode(val & mask))
            val >>= t.width

        return cls(ret)

    @class_and_instance_method
    def copy(self):
        type(self)(self)



"""Implements fixed width integer types: :class:`Uint` - unsigned, :class:`Int`
- signed and :class:`Integer` - sign agnostic type. These types correspond to
HDL logic vector types.

Objects of these classes can also be instantiated and they provide some integer
arithmetic capabilities.
"""

from .base import class_and_instance_method
from .base import typeof, EnumerableGenericMeta, is_type
from .number import Number
from .math import bitw
from .unit import Unit
from functools import reduce
from operator import matmul


def concat(l):
    return reduce(matmul, l)


class IntegralType(EnumerableGenericMeta):
    def __new__(cls, name, bases, namespace, args=None):
        if args is not None:
            err = None
            try:
                if isinstance(args, tuple):
                    args = tuple(a if isinstance(a, (str, bytes)) else int(a) for a in args)
                elif isinstance(args, dict):
                    args = {
                        n: a if isinstance(a, (str, bytes)) else int(a)
                        for n, a in args.items()
                    }
                else:
                    args = args if isinstance(args, (str, bytes)) else int(args)
            except TypeError as e:
                err = e

            if err:
                raise TypeError(f"{cls} type parameters must be all either integers or strings,"
                                f"not '{args}'")

        return super().__new__(cls, name, bases, namespace, args)

    def __getitem__(self, index):
        if not self.specified:
            return super().__getitem__(index)

        index = self.index_norm(index)

        width = 0
        for i in index:
            if isinstance(i, slice):
                if i.stop == 0 or i.start >= len(self):
                    return Unit
                elif i.stop > len(self):
                    i.stop = len(self)

                width += i.stop - i.start
            else:
                if i >= len(self):
                    raise IndexError
                width += 1

        return Uint[width]

    @property
    def mask(self) -> int:
        return (1 << self.width) - 1

    @property
    def width(self) -> int:
        return self.__args__[-1]

    def keys(self):
        """Returns a list of keys that can be used for indexing the type.

        >>> Int[8].keys()
        [0, 1, 2, 3, 4, 5, 6, 7]
        """
        return list(range(self.width))


class Integral(int, metaclass=IntegralType):
    def __hash__(self):
        return hash((type(self), int(self)))


Number.register(Integral)


# TODO: Should this implement __eq__ and __ne__?

class IntegerType(IntegralType):
    """Defines lib methods for all Integer based classes.
    """
    def __new__(cls, name, bases, namespace, args=None):
        if args:
            if isinstance(args, dict):
                w = list(args.values())[0]
            elif isinstance(args, (tuple, list)):
                w = args[0]
            else:
                w = args

            if not isinstance(w, (str, int)):
                raise TypeError(f"{name} type parameter must be an integer, not '{w}'")

            if isinstance(w, int):
                if w < 0:
                    raise TypeError(f"{name} type parameter must be a positive integer, not '{w}'")

                args = (int(w), )

        return super().__new__(cls, name, bases, namespace, args=args)

    def __abs__(self):
        if not self.signed:
            return self

        return Int[self.width + 1]

    def __add__(self, other):
        if not typeof(other, Integer):
            return NotImplemented

        signed = self.signed or other.signed

        w1 = self.width + 1 if signed and not self.signed else self.width
        w2 = other.width + 1 if signed and not other.signed else other.width
        res_type = Int if signed else Uint

        return res_type[max((w1, w2)) + 1]

    def __and__(self, other):
        return self.base[max(op.width for op in (self, other))]

    def __ceil__(self):
        return self

    def __float__(self):
        return float

    def __floordiv__(self, other):
        return self.base[self.width - other.width + 1]

    def __floor__(self):
        return self

    def __ge__(self, other):
        return self.width >= other.width

    def __gt__(self, other):
        return self.width > other.width

    def __iadd__(self, other):
        return self

    def __imul__(self, other):
        return self

    def __int__(self):
        return int

    def __invert__(self):
        return self

    def __isub__(self, other):
        return self

    def __le__(self, other):
        return Bool

    def __lt__(self, other):
        return Bool

    def __lshift__(self, other):
        return self.base[self.width + int(other)]

    def __mod__(self, other):
        return other

    def __mul__(self, other):
        """Returns the same type, whose width is equal to the sum of operand widths
        if both operands are unsigned.
        Returns the signed Int type if other operand is signed.

        >>> Uint[8] + Uint[8]
        Uint[16]
        >>> Uint[8] + Int[8]
        Int[16]
        """

        if not typeof(other, Integer):
            return NotImplemented

        ops = [self, other]

        signed = any(typeof(op, Int) for op in ops)
        if signed:
            return Int[self.width + other.width]
        else:
            return self.base[self.width + other.width]

    def __neg__(self):
        return Int[self.width + 1]

    def __or__(self, other):
        # return self.width | other.width
        return self.base[max(op.width for op in (self, other))]

    __radd__ = __add__

    def __rfloordiv__(self, other):
        return self.base[other.width - self.width + 1]

    def __rtruediv__(self, other):
        return self.base[other.width - self.width + 1]

    def __rmod__(self, other):
        return self

    __rmul__ = __mul__

    def __rshift__(self, other):
        shamt = int(other)
        width = len(self)

        if shamt > width:
            return NotImplemented

        if shamt == width:
            return Unit
        else:
            return self.base[width - shamt]

    def __str__(self):
        if self.args:
            if isinstance(self.args[0], int):
                return f'z{self.args[0]}'
            else:
                return f'z({self.args[0]})'
        else:
            return super().__str__()

    def __sub__(self, other):
        if not typeof(other, Integer):
            return NotImplemented

        signed = self.signed or other.signed

        w1 = self.width + 1 if signed and not self.signed else self.width
        w2 = other.width + 1 if signed and not other.signed else other.width

        return Int[max((w1, w2)) + 1]

    def __truediv__(self, other):
        return self.base[self.width - other.width + 1]

    def __xor__(self, other):
        return self.base[max(op.width for op in (self, other))]

    @property
    def specified(self):
        return False


def check_width(val, width, type_):
    if ((bitw(val) > width) and (val != 0)):
        raise ValueError(f"{repr(type_)} cannot represent value '{val}'")


class Integer(Integral, metaclass=IntegerType):
    """Base type for both :class:`Int` and :class:`Uint` generic types.
    Corresponds to HDL logic vector types. For an example Integer[9] translates
    to :sv:`logic [8:0]`.
    """
    def __new__(cls, val: int = None):
        # TODO: Instantiating a type that has templated arguments unresolved
        # should throw a nicer error (same for all other types)
        if val is None:
            val = 0

        if type(val) == cls:
            return val

        if not typeof(type(val), Integer) and not isinstance(val, (int, list, float)):
            if is_type(type(val)):
                raise ValueError(f"cannot convert '{val}' of type '{repr(type(val))}' to '{cls}'")

            try:
                val = int(val)
            except:
                raise ValueError(f"cannot convert '{val}' of type '{repr(type(val))}' to '{cls}'")

        if isinstance(val, list):
            if cls.signed:
                raise TypeError(f'cannot create Int from bit list')

            if not cls.specified:
                cls = Uint[len(val)]

            ival = 0
            for v in val:
                ival <<= 1
                ival |= bool(v)

            return cls(ival)

        if cls.base is Integer:
            if typeof(type(val), Uint) or int(val) >= 0:
                if cls.is_generic():
                    if typeof(type(val), Uint):
                        cls = type(val)
                    else:
                        cls = Uint[bitw(val)]
                else:
                    cls = Uint[cls.width]

            if typeof(type(val), Int) or int(val) < 0:
                if cls.is_generic():
                    if typeof(type(val), Int):
                        cls = type(val)
                    else:
                        cls = Int[bitw(val)]
                else:
                    cls = Int[cls.width]

        if typeof(cls, Uint) and val < 0:
            raise ValueError(f"cannot represent negative numbers with unsigned type '{repr(cls)}'")

        if cls.is_generic():
            res = cls[bitw(val)](int(val))
        else:
            res = super(Integer, cls).__new__(cls, val)

        check_width(val, res.width, cls)
        return res

    def __abs__(self):
        if self.signed:
            return abs(type(self))(-int(self) if code(self, Uint)[-1] else int(self))
        else:
            return abs(type(self))(int(self))

    def __add__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if isinstance(other, Integer):
            conv_other = other
        elif isinstance(other, (int, float)):
            conv_other = Integer(other)
        else:
            return NotImplemented

        return (type(self) + type(conv_other))(super().__add__(conv_other))

    def __eq__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if isinstance(other, Integer):
            conv_other = other
        elif isinstance(other, (int, float)):
            conv_other = Integer(other)
        else:
            return NotImplemented

        return Bool(super().__eq__(conv_other))

    @class_and_instance_method
    def __getitem__(self, index):
        """Returns the value of the indexth bit in the number representation.

        >>> Integer[8](0b10101010)[5]
        1

        >>> Integer[8](0b10101010)[1::2]
        Uint[4](15)
        """

        index = type(self).index_norm(index)

        base = None
        for i in index:
            if isinstance(i, slice):
                start, stop, _ = i.indices(self.width)

                if stop <= start:
                    part = Unit()
                else:
                    part = Uint[stop - start]((int(self) & ((1 << stop) - 1)) >> start)

            elif i < self.width:
                part = Bool(int(self) & (1 << i))
            else:
                raise IndexError

            if base is None:
                base = part
            else:
                base = part @ base

        return base

    def __hash__(self):
        return super().__hash__()

    def __iadd__(self, other):
        if not is_type(type(other)):
            other = type(self).base(other)

        if not isinstance(other, Integer):
            return NotImplemented

        if not self.signed and other.signed:
            raise TypeError(
                f"unsupported operand type(s) for +=: '{type(self)}' and '{type(other)}'")

        return type(self)(int(self) + int(other))

    def __imul__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if isinstance(other, Integer):
            conv_other = other
        else:
            conv_other = Integer(other)

        return type(self)(super().__mul__(conv_other))


    def __invert__(self):
        return type(self)(super().__invert__() & type(self).mask)

    def __isub__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if is_type(type(other)):
            conv_other = other
        elif isinstance(other, (int, float)):
            conv_other = Integer(other)
        else:
            return NotImplemented

        return type(self)(super().__sub__(conv_other))

    def __lshift__(self, other):
        return (type(self) << other)(super().__lshift__(other))

    def __len__(self):
        """Returns the number of bits used for the representation

        >>> len(Integer[8](0))
        8
        """
        return len(type(self))

    def __mul__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if isinstance(other, Integer):
            conv_other = other
        else:
            conv_other = Integer(other)

        return (type(self) * type(conv_other))(super().__mul__(conv_other))

    def __ne__(self, other):
        if not is_type(type(other)):
            return super().__ne__(other)

        return not self.__eq__(other)

    def __neg__(self):
        return (-type(self)).decode(super().__neg__())

    def __repr__(self):
        return f'{repr(type(self))}({int(self)})'

    __rmul__ = __mul__

    def __rshift__(self, other):
        if typeof((type(self) >> other), Unit):
            return Unit()

        return (type(self) >> other)(super().__rshift__(other))

    def __rsub__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if is_type(type(other)):
            conv_other = other
        elif isinstance(other, (int, float)):
            conv_other = Integer(other)
        else:
            return NotImplemented

        return (type(conv_other) - type(self))(super().__rsub__(conv_other))

    def __str__(self):
        return f'{str(type(self))}({int(self)})'

    def __sub__(self, other):
        if is_type(type(other)) and not isinstance(other, Integer):
            return NotImplemented

        if is_type(type(other)):
            conv_other = other
        elif isinstance(other, (int, float)):
            conv_other = Integer(other)
        else:
            return NotImplemented

        return (type(self) - type(conv_other))(super().__sub__(conv_other))

    @property
    def quant(self):
        return self.decode(1)

    @class_and_instance_method
    @property
    def width(self):
        """Returns the number of bits used for the representation

        >>> Integer[8](0).width
        8
        """
        return type(self).width

    @class_and_instance_method
    @property
    def mask(self):
        return (1 << self.width) - 1

    def code(self):
        return int(self) & self.mask

    @classmethod
    def decode(cls, val):
        """Creates Integer object from any int-convertible object val.

        >>> Integer[8].decode(0xffff)
        Integer[8](255)
        """
        return cls(int(val))


class IntType(IntegerType):
    def __str__(self):
        if self.args:
            if isinstance(self.args[0], int):
                return f'i{self.args[0]}'
            else:
                return f'i({self.args[0]})'
        else:
            return super().__str__()

    @property
    def max(self):
        return self.decode(2**(self.width - 1) - 1)

    @property
    def min(self):
        return self.decode(-2**(self.width - 1))

    @property
    def specified(self):
        return IntegralType.specified.fget(self)

    @property
    def signed(self):
        return True


class Int(Integer, metaclass=IntType):
    """Fixed width generic signed integer data type.

    Generic parameters:
       N: Bit width of the :class:`Int` representation

    Args:
       val: Integer value to convert to :class:`Int`

    :class:`Int` is a generic datatype derived from :class:`Integer`. It
    represents signed integers with fixed width binary representation. Concrete
    data type is obtained by indexing:

    >>> i16 = Int[16]

    """
    __parameters__ = ['N']

    @class_and_instance_method
    @property
    def signed(self):
        return True

    def is_abstract(self):
        return False

    def __new__(cls, val: int = 0):
        if type(val) == cls:
            return val

        if cls.is_generic():
            if isinstance(val, Int):
                return val
            elif isinstance(val, Uint):
                res = cls[val.width + 1](int(val))
            else:
                res = cls[bitw(val) + (0 if int(val) < 0 else 1)](int(val))
        else:
            res = super(Int, cls).__new__(cls, val)
            # res = super(Int, cls).__new__(cls,
            #                               int(val) & ((1 << len(cls)) - 1))

        check_width(val, res.width if val < 0 else res.width - 1, cls)
        return res

    @classmethod
    def decode(cls, val):
        val = int(val & ((1 << cls.width) - 1))
        if val >= (1 << (cls.width - 1)):
            val -= 1 << cls.width

        return cls(val)


class UintType(IntegerType):
    """Fixed width generic unsigned integer data type.

    Generic parameters:
       N: Bit width of the :class:`Uint` representation

    :class:`Uint` is a generic datatype derived from :class:`Integer`. It
    represents unsigned integers with fixed width binary representation.
    Concrete data type is obtained by indexing:

    >>> u16 = Uint[16]

    """
    @property
    def signed(self):
        return False

    def is_abstract(self):
        return False

    @property
    def specified(self):
        return IntegralType.specified.fget(self)

    def __matmul__(self, other):
        if not typeof(other, (bool, Uint)):
            return NotImplemented

        if issubclass(other, bool):
            return Uint[self.width + 1]
        else:
            return Uint[self.width + other.width]

    @property
    def max(self):
        return self.decode(2**self.width - 1)

    @property
    def min(self):
        return self.decode(0)

    def __str__(self):
        if not self.args:
            return f'u'
        elif isinstance(self.args[0], int):
            return f'u{self.args[0]}'
        else:
            return f'u({self.args[0]})'


class Uint(Integer, metaclass=UintType):
    """Implements the :class:`Uint` type instance.

    Args:
       val: Integer value to convert to :class:`Uint`

    :class:`Uint` is a generic datatype derived from :class:`Integer`. It
    represents unsigned integers with fixed width binary representation.

    >>> Uint[16](0xffff)
    Uint[16](65535)

    """
    __parameters__ = ['N']

    @class_and_instance_method
    @property
    def signed(self):
        return False

    def __matmul__(self, other):
        if isinstance(other, bool):
            return Uint[self.width + 1]((int(self) << 1) | int(other))

        if not is_type(type(other)):
            raise TypeError(
                f"unsupported operand type(s) for @: '{type(self)}' and '{type(other)}'")

        if isinstance(other, Unit):
            return self

        if not isinstance(other, Uint):
            other = Uint(other)

        return Uint[self.width + other.width]((int(self) << other.width) | int(other))

    def __rmatmul__(self, other):
        if isinstance(other, bool):
            return Bool(other) @ self

        if isinstance(other, Unit):
            return self

        return NotImplemented


class BoolMeta(UintType):
    def __new__(cls, name, bases, namespace, args=None):
        spec_cls = super().__new__(cls, name, bases, namespace, args=(1, ))
        spec_cls._base = Uint
        return spec_cls

    def __repr__(self):
        return 'Bool'

    def copy(self):
        return self


class Bool(Uint, metaclass=BoolMeta):
    __parameters__ = ['N']

    def __new__(cls, val=None):
        return int.__new__(cls, bool(val))


def code(data, cast_type=Uint):
    if is_type(data):
        if is_type(cast_type) and not cast_type.specified and typeof(cast_type, Integer):
            cast_type = cast_type[data.width]

        return cast_type

    dtype = type(data)
    if is_type(dtype):
        data = data.code()

        if is_type(cast_type) and not cast_type.specified and typeof(cast_type, Integer):
            cast_type = cast_type[dtype.width]

    if is_type(cast_type) and cast_type.specified:
        return cast_type.decode(data & ((1 << cast_type.width) - 1))
    elif is_type(type(data)):
        return cast_type(data.code())
    else:
        return cast_type(data)


def decode(val):
    if is_type(type(val)):
        return val.decode()
    else:
        return val



from pygears import gear, datagear, alternative, module
from pygears.typing.qround import get_out_type, get_cut_bits
from pygears.typing import Uint, code, Bool, Int, Fixp, Ufixp


@datagear
def qround(din,
           *,
           fract=0,
           cut_bits=b'get_cut_bits(din, fract)',
           signed=b'din.signed') -> b'get_out_type(din, fract)':

    res = code(din, Int if signed else Uint) + (Bool(1) << (cut_bits - 1))
    return code(res >> cut_bits, module().tout)


# @datagear
# def qround_even(din,
#                 *,
#                 fract=0,
#                 cut_bits=b'get_cut_bits(din, fract)',
#                 signed=b'din.signed') -> b'get_out_type(din, fract)':

#     val_coded = code(din, Int if signed else Uint)
#     round_bit = val_coded[cut_bits]

#     res = val_coded + Uint([round_bit] + [~round_bit] * (cut_bits - 1))
#     return code(res[cut_bits:])


@gear
async def truncate(din, *, nbits=2) -> b'din':
    pass


@gear
async def round_half_up(din, *, nbits=2) -> b'din':
    pass


@gear
async def round_to_zero(din, *, nbits=2) -> b'din':
    pass


@gear
async def round_to_even(din, *, nbits=2) -> b'din':
    async with din as d:
        return round(float(d) / (2**nbits)) * (2**nbits)



from pygears import gear
from pygears.typing import Queue


@gear
async def qinterlace(*din: Queue['t_data']) -> b'Queue[t_data, 2]':
    """Short for Trasaction Round Robin, outputs data from one of the input
    interfaces following a `Round Robin` schedule.

    Returns:
        A level 2 :class:`Queue` type where the higher level signals that one
          round is finished. The lower level is passed from input.
    """
    for i, data_in in enumerate(din):
        async for (data, eot) in data_in:
            yield (data, (i == len(din) - 1) @ eot)


@gear
async def interlace(*din: b'din_t') -> b'din_t':
    for i, data_in in enumerate(din):
        async with data_in as data:
            yield data



"""This module implements various logging facilities for the PyGears framework.
It is a wrapper around standard Python `logging
<https://docs.python.org/library/logging.html>`__ but provides some additional
features like:

- automatic exception raising when logging at any level
- customized stack trace printing
- logging to temporary files
- integration with PyGears registry for configuration

To register a new logger create a :class:`CustomLog` instance by specifying the
logger name and the default logging level:

>>> CustomLog('core', log.WARNING)

.. _levels:

Logging Levels
--------------

The numeric values of logging levels are given in the following table.

+--------------+---------------+
| Level        | Numeric value |
+==============+===============+
| ``CRITICAL`` | 50            |
+--------------+---------------+
| ``ERROR``    | 40            |
+--------------+---------------+
| ``WARNING``  | 30            |
+--------------+---------------+
| ``INFO``     | 20            |
+--------------+---------------+
| ``DEBUG``    | 10            |
+--------------+---------------+
| ``NOTSET``   | 0             |
+--------------+---------------+

"""

import copy
import logging
import os
import sys
import tempfile
from functools import partial
from logging import CRITICAL, DEBUG, ERROR, INFO, NOTSET, WARNING

from .registry import Inject, PluginBase, inject, reg
from .trace_format import enum_stacktrace

HOOKABLE_LOG_METHODS = ['critical', 'error', 'warning', 'info', 'debug']


class LogException(Exception):
    def __init__(self, message, name):
        super().__init__(message)
        self.name = name


def set_log_level(var, level, name):
    log = logging.getLogger(name)
    if log.level != level:
        log.setLevel(level)
        for h in log.handlers:
            h.setLevel(level)


def log_parent(cls, severity):
    def wrapper(self, msg, *args, **kwargs):
        '''Wrapper around logger methods from HOOKABLE_LOG_METHODS
        for calling hooks'''
        getattr(super(CustomLogger, self), severity)(msg, *args, **kwargs)
        getattr(self, f'{severity}_hook')(logger_name=self.name, message=msg)

    return wrapper


def log_action_exception(name, message):
    raise LogException(message, name)


def log_action_debug():
    import pdb
    pdb.set_trace()


@inject
def custom_action(logger_name, message, severity, log_cfgs=Inject('logger')):
    method = reg[f'logger/{logger_name}/{severity}']
    if method == 'exception':
        log_action_exception(logger_name, message)
    elif method == 'debug':
        log_action_debug()
    elif method == 'pass':
        pass
    elif callable(method):
        # custom function in registry
        method(message)


def log_hook(cls, severity):
    @inject
    def wrapper(self, logger_name, message, hooks=Inject('logger/hooks')):
        '''Custom <severity>_hook methods for HOOKABLE_LOG_METHODS.'''
        custom_action(logger_name, message, severity)
        for hook in hooks:
            hook(logger_name, severity, message)

    return wrapper


def hookable_methods_gen(cls):
    for severity in HOOKABLE_LOG_METHODS:
        setattr(cls, severity, log_parent(cls, severity))
        setattr(cls, f'{severity}_hook', log_hook(cls, severity))
    return cls


class LogFmtFilter(logging.Filter):
    @inject
    def __init__(self,
                 name='',
                 stack_traceback_fn=Inject('logger/stack_traceback_fn')):
        super(LogFmtFilter, self).__init__(name)
        self.stack_traceback_fn = stack_traceback_fn

    def filter(self, record):
        record.stack_file = ''
        record.err_file = ''

        if record.levelno > INFO:
            if os.path.exists(self.stack_traceback_fn):
                with open(self.stack_traceback_fn) as f:
                    stack_num = sum(1 for _ in f)
            else:
                stack_num = 0

            # TODO: improve error reporting in this way
            record.stack_file = f'\n  File "{self.stack_traceback_fn}", line {stack_num}, for stacktrace'
            *_, last_frame = enum_stacktrace()
            record.err_file = f'\n{last_frame}' [:-1]

        return True


@hookable_methods_gen
class CustomLogger(logging.Logger):
    '''Inherits from Logger and adds hook methods to HOOKABLE_LOG_METHODS'''
    def __init__(self, name, level=INFO):
        super(CustomLogger, self).__init__(name, level)

    def get_format(self):
        return logging.Formatter(
            '%(name)s [%(levelname)s]: %(message)s %(err_file)s %(stack_file)s'
        )

    def get_filter(self):
        return LogFmtFilter()

    def get_logger_handler(self, handler=None):
        if handler is None:
            handler = logging.StreamHandler(sys.stdout)

        handler.setLevel(self.level)
        handler.setFormatter(self.get_format())

        filt = self.get_filter()
        if filt is not None:
            handler.addFilter(filt)

        return handler


def core_log():
    return logging.getLogger('core')


def typing_log():
    return logging.getLogger('typing')


def util_log():
    return logging.getLogger('util')


def gear_log():
    return logging.getLogger('gear')


def conf_log():
    return logging.getLogger('conf')


def register_custom_log(name, level=INFO, cls=CustomLogger):
    '''PyGears integrated logger class.

    Args:
        name: logger name
        verbosity: default logging level:

    CustomLog instances are customizable via :samp:`logger/{logger_name}`
    :ref:`registry <registry:registry>` subtree. The logger instance registry subtree
    contains the following configuration variables:

    - ``level`` (int): All messages that are logged with a verbosity level
      below this configured ``level`` value will be discarded. See
      :ref:`levels` for a list of levels.
    - ``print_traceback`` (bool): If set to ``True``, the traceback will be
      printed along with the log message.

    - optional level name with desired action. Custom actions can be set for
      any verbosity level by passing the function or any already supported
      action to the appropriate registry subtree. Supported actions are:

      - ``exception``: if set, an exception will be raised whenever logging the
        message at the desired level
      - ``debug``: if set, the debugger will be started and execution paused
      - ``pass``: if set, the message will be printed and no further action
        taken; this is usefull for clearing prevously set values

    Sets the verbosity level for the ``core`` logger at ``INFO`` level:

    >>> reg['logger/core/level'] = INFO

    Configures the ``typing`` logger to throw exception on warnings:

    >>> reg['logger/typing/warning'] = 'exception'

    Configures the ``conf`` logger to use custom function on errors:

    >>> reg['logger/conf/errors'] = custom_func
    '''
    log_cls = logging.getLoggerClass()
    logging.setLoggerClass(cls)

    reg_name = f'logger/{name}'

    for m in HOOKABLE_LOG_METHODS:
        reg.confdef(f'{reg_name}/{m}', default='pass')

    reg.confdef(f'{reg_name}/level',
                  default=level,
                  setter=partial(set_log_level, name=name))

    reg.confdef(f'{reg_name}/print_traceback', default=True)

    logger = logging.getLogger(name)
    logger.handlers.clear()
    logger.setLevel(level)
    logger.addHandler(logger.get_logger_handler())

    logging.setLoggerClass(log_cls)


class LogPlugin(PluginBase):
    @classmethod
    def bind(cls):
        tf = tempfile.NamedTemporaryFile(delete=False)
        reg['logger/stack_traceback_fn'] = tf.name
        reg['logger/hooks'] = []

        register_custom_log('core', WARNING)
        register_custom_log('typing', WARNING)
        register_custom_log('util', WARNING)
        register_custom_log('gear', WARNING)
        register_custom_log('conf', WARNING)



import asyncio
import itertools
import logging
import os
import random
import sys
import tempfile
import threading
import time
import stopit
import contextlib

from functools import partial
from pygears import GearDone, find, reg
from pygears.conf import CustomLogger, LogFmtFilter, register_custom_log
from pygears.core.gear import GearPlugin, Gear
# from pygears.core.intf import get_consumer_tree as intf_get_consumer_tree
from pygears.core.port import InPort, OutPort, HDLConsumer, HDLProducer
from pygears.core.sim_event import SimEvent
from pygears.core.hier_node import HierVisitorBase
from pygears.util.fileio import expand
from pygears.sim import log

gear_reg = {}
sim_reg = {}


class SimFinish(Exception):
    pass


class SimCyclic(Exception):
    pass


class SimTimeout(Exception):
    pass


def timestep():
    try:
        return sim_reg['timestep']
    except KeyError:
        return None


def sim_phase():
    return sim_reg['simulator'].phase


def clk():
    # TODO: await clk() after yield (in back phase) might not have desired effect
    gear_reg['current_sim'].phase = 'forward'
    return sim_reg['clk_event'].wait()


async def delta():
    gear_reg['current_sim'].phase = 'back'
    await asyncio.sleep(0)
    return sim_reg['simulator'].phase


def artifacts_dir():
    return reg['results-dir']


def schedule_to_finish(gear):
    sim = sim_reg['simulator']
    if gear in sim_reg['map']:
        sim.schedule_to_finish(sim_reg['map'][gear])


class SimFuture(asyncio.Future):
    def coro_iter(self):
        yield self

    def __iter__(self):
        return self.coro_iter()

    __await__ = __iter__


# A recursive function used by topo_sort
def topo_sort_util(v, g, dag, visited, stack, cycle):

    # Mark the current node as visited.
    visited[v] = True
    cycle.append(g)

    # Recur for all the vertices adjacent to this vertex
    for consumer in dag[g]:
        i = list(dag.keys()).index(consumer)
        if consumer in cycle:
            index = cycle.index(consumer)
            cycle.append(consumer)
            raise SimCyclic('Simulation not possible, gear cycle found:'
                            f' {" -> ".join([c.name for c in cycle[index:]])}')

        if not visited[i]:
            topo_sort_util(i, consumer, dag, visited, stack, cycle)

    cycle.pop()

    # Push current vertex to stack which stores result
    stack.insert(0, g)


def topo_sort(dag):
    # Mark all the vertices as not visited
    visited = [False] * len(dag)
    stack = []

    # for i, g in enumerate(dag):
    #     if not g.in_ports:
    #         stack.append(g)
    #         visited[i] = True

    # Call the recursive helper function to store Topological
    # Sort starting from all vertices one by one
    for i, g in enumerate(dag):
        if not visited[i]:
            topo_sort_util(i, g, dag, visited, stack, [])

    return stack


def _get_consumer_tree_rec(root_intf, cur_intf, consumers):
    for port in cur_intf.consumers:
        if isinstance(port, HDLConsumer):
            continue

        cons_intf = port.consumer
        if port in reg['sim/map']:
            consumers.append(port)
        elif port.gear.hierarchical:
            if cons_intf is None:
                raise Exception(f'Port {port.name} found dangling')

            _get_consumer_tree_rec(root_intf, cons_intf, consumers)
        else:
            consumers.append(port.gear)


def get_consumer_tree(intf):
    consumers = []
    _get_consumer_tree_rec(intf, intf, consumers)
    return consumers


class GearEnum(HierVisitorBase):
    def __init__(self):
        self.gears = []

    def Gear(self, node):
        if not node.hierarchical:
            self.gears.append(node)


def cosim_build_dir(top):
    from pygears.core.gear_memoize import gear_inst_hash
    hsh = gear_inst_hash(top)
    return f'{top.basename}_{hex(hsh & 0xfffffff)}'


def cosim(top, sim, *args, **kwds):
    if top is None:
        top = reg['gear/root']
    elif isinstance(top, str):
        top_name = top
        top = find(top)

        if top is None:
            raise Exception(f'No gear found on path: "{top_name}"')

    kwds['outdir'] = expand(kwds.get('outdir', reg['results-dir']))
    kwds['rebuild'] = kwds.get('rebuild', True)

    reg['sim/hook/cosim_build_before'](top, args, kwds)

    if not kwds['rebuild']:
        import sys
        if sys.flags.hash_randomization:
            raise Exception(
                f"Reuse of previous cosim build turned on for module '{top}'.\n"
                " In order to use this feature you need to turn off Python hash randomization.\n"
                " Before running Python, you need to set environment variable 'PYTHONHASHSEED=0'")

    if isinstance(sim, str):
        if sim in ['cadence', 'xsim', 'questa']:
            from .modules import SimSocket
            sim_cls = SimSocket
            kwds['sim'] = sim
        elif sim == 'verilator':
            from .modules import SimVerilated
            from .modules.verilator import build

            timeout = kwds.pop('timeout', 100)
            sim_cls = SimVerilated
            build(top, **kwds)
            kwds['timeout'] = timeout
        else:
            raise Exception(f"Unsupported simulator: {sim}")

        kwds['rebuild'] = False
    else:
        sim_cls = sim

    # TODO: Should we invoke build here even when simulation class is sent
    # directly
    reg['sim/hook/cosim_build_after'](top, args, kwds)

    if args or kwds:
        top.params['sim_cls'] = partial(sim_cls, *args, **kwds)
    else:
        top.params['sim_cls'] = sim_cls


def simgear_exec_order(gears):
    sim_map = reg['sim/map']
    dag = {}

    for g in gears:
        dag[g] = []
        for p in g.out_ports:
            dag[g].extend(get_consumer_tree(p.consumer))

    for g, sim_gear in sim_map.items():
        if isinstance(g, OutPort):
            if (not g.gear.hierarchical):
                dag[g.gear].clear()

    for g, sim_gear in sim_map.items():
        if isinstance(g, InPort):
            # TODO: Following doesn't work when verilator adds InputPorts for channeled intfs
            # if (len(g.consumer.consumers) == 1
            #         and isinstance(g.consumer.consumers[0], HDLConsumer)):

            if (not g.gear.hierarchical):
                dag[g] = [g.gear]
            else:
                dag[g] = get_consumer_tree(g.consumer)

        elif isinstance(g, OutPort):
            # TODO: Test if this works
            # from pygers.core.graph import has_async_producer
            # if has_async_producer(g):
            if (not g.gear.hierarchical):
                dag[g.gear].append(g)

            dag[g] = get_consumer_tree(g.consumer)

    gear_order = topo_sort(dag)

    cosim_modules = [g for g in sim_map if isinstance(g, Gear) and g.params['sim_cls'] is not None]

    gear_multi_order = cosim_modules.copy()
    for g in gear_order:
        if (all(not m.has_descendent(g) for m in cosim_modules) or isinstance(g,
                                                                              (InPort, OutPort))):
            gear_multi_order.append(g)

    # print('Order: ')
    # for g in gear_multi_order:
    #     print(g.name)

    # import networkx as nx
    # import matplotlib.pyplot as plt

    # G = nx.DiGraph()
    # for g in dag:
    #     for c in dag[g]:
    #         G.add_edge(g.name, c.name)

    # pos = nx.spring_layout(G)
    # nx.draw(G, pos, font_size=16, with_labels=False)

    # for p in pos:  # raise text positions
    #     pos[p][1] += 0.07

    # nx.draw_networkx_labels(G, pos)
    # plt.show()

    return gear_multi_order


class EventLoop(asyncio.events.AbstractEventLoop):
    def __init__(self, step_timeout=2):
        self.step_timeout = step_timeout
        self.events = {
            'before_setup': SimEvent(),
            'before_run': SimEvent(),
            'after_run': SimEvent(),
            'before_call_forward': SimEvent(),
            'after_call_forward': SimEvent(),
            'before_call_back': SimEvent(),
            'after_call_back': SimEvent(),
            'before_timestep': SimEvent(),
            'after_timestep': SimEvent(),
            'after_cleanup': SimEvent(),
            'after_finish': SimEvent(),
            'at_exit': SimEvent()
        }

    def get_debug(self):
        return False

    def insert_gears(self, gears, pos=None):
        for g in gears:
            g.phase = 'forward'
            if g not in self.sim_map:
                raise Exception(f'Gear "{g.name}" of type "{g.definition.__name__}" has'
                                f' no simulation model')

            self.sim_map[g].phase = 'forward'

        sim_gears = [self.sim_map[g] for g in gears]

        if pos is None:
            index = 0
        elif isinstance(pos, int):
            index = pos
        else:
            index = self.sim_gears.index(self.sim_map[pos])

        self.sim_gears[index:index] = sim_gears

        for g in set(sim_gears):
            self.tasks[g] = g.run()
            self.task_data[g] = None

    def future_done(self, fut):
        # This will never be called, call_soon will be called first and handle the finishing
        pass

    def call_soon(self, callback, fut, context=None):
        sim_gear = self.wait_list.pop(fut, None)
        if fut.cancelled():
            # Interface that sim_gear waited on is done, so we need to finish
            # the sim_gear too
            if sim_gear:
                self.schedule_to_finish(sim_gear)
        else:
            # If sim_gear was waiting for ack
            if sim_gear.phase == 'back':
                self.back_ready.add(sim_gear)
            else:
                self.forward_ready.add(sim_gear)

    def remove(self, sim_gear):
        # log.info(f'Remove: {sim_gear.port.name if hasattr(sim_gear, "port") else sim_gear.gear.name}')
        self.sim_map.pop(sim_gear.gear, None)
        self.tasks.pop(sim_gear, None)
        self.task_data.pop(sim_gear, None)

        if sim_gear.parent is None:
            self.sim_gears.remove(sim_gear)

        self.back_ready.discard(sim_gear)
        self.forward_ready.discard(sim_gear)
        self.back_ready.discard(sim_gear)
        self.forward_ready.discard(sim_gear)

    def create_future(self):
        """Create a Future object attached to the loop."""
        # fut = SimFuture()
        fut = asyncio.Future()
        fut.add_done_callback(self.future_done)

        return fut

    def schedule_to_finish(self, sim_gear):
        '''Schedule module to be finished during next back phase.'''
        self._schedule_to_finish.add(sim_gear)

    def _finish(self, sim_gear):
        if sim_gear.done:
            return

        try:
            self.cur_gear = sim_gear.gear
            reg['gear/current_module'] = self.cur_gear
            reg['gear/current_sim'] = sim_gear
            self.tasks[sim_gear].throw(GearDone)
        except (StopIteration, GearDone):
            pass
        except Exception as e:
            raise e
        else:
            log.error(
                f"Process of '{sim_gear.port.name if hasattr(sim_gear, 'port') else sim_gear.gear.name}'"
                f" didn't stop on finish!")
        finally:
            self.events['after_finish'](self, sim_gear)
            self.cur_gear = reg['gear/root']
            reg['gear/current_module'] = self.cur_gear
            reg['gear/current_sim'] = sim_gear
            self.remove(sim_gear)

    def run_gear(self, sim_gear, ready):
        before_event = None
        after_event = None

        if ready is self.forward_ready:
            before_event = self.events['before_call_forward']
            after_event = self.events['after_call_forward']
        elif ready is self.back_ready:
            before_event = self.events['before_call_back']
            after_event = self.events['after_call_back']

        if before_event:
            before_event(self, sim_gear)

        try:
            data = self.tasks[sim_gear].send(self.task_data[sim_gear])
        except (StopIteration, GearDone):
            self.done.add(sim_gear)
        else:
            self.wait_list[data] = sim_gear
            if isinstance(data, asyncio.Future):
                self.wait_list[data] = sim_gear
            else:
                if sim_gear.phase == 'back':
                    self.back_ready.add(sim_gear)
                else:
                    self.forward_ready.add(sim_gear)

            self.task_data[sim_gear] = data
        finally:
            if after_event:
                after_event(self, sim_gear)

    def maybe_run_gear(self, sim_gear, ready):
        ready.discard(sim_gear)
        # self.delta_ready.discard(sim_gear)

        self.cur_gear = sim_gear.gear
        gear_reg['current_module'] = self.cur_gear
        gear_reg['current_sim'] = sim_gear

        # print(
        #     f'{self.phase}: '
        #     f'{sim_gear.port.name if hasattr(sim_gear, "port") else sim_gear.gear.name}'
        # )
        self.run_gear(sim_gear, ready)

        self.cur_gear = gear_reg['root']
        gear_reg['current_module'] = self.cur_gear
        gear_reg['current_sim'] = sim_gear

    def sim_list(self, sim_gears):
        if self.phase == 'forward':
            ready_dict = self.forward_ready
        else:
            sim_gears = reversed(sim_gears)
            ready_dict = self.back_ready

        for sim_gear in sim_gears:
            if sim_gear in ready_dict:
                self.maybe_run_gear(sim_gear, ready_dict)

            if sim_gear.child:
                unprocessed_children = True
                while (unprocessed_children):
                    self.sim_list(sim_gear.child)

                    cur_child_num = len(sim_gear.child)
                    if sim_gear in ready_dict:
                        self.maybe_run_gear(sim_gear, ready_dict)

                    unprocessed_children = cur_child_num < len(sim_gear.child)

    def sim_loop(self, timeout):
        clk = reg['sim/clk_event']
        delta = reg['sim/delta_event']

        global gear_reg, sim_reg
        gear_reg = reg['gear']._dict
        sim_reg = reg['sim']._dict

        reg['sim/timestep'] = 0

        timestep = -1
        start_time = time.time()
        finished = False

        log.info("-------------- Simulation start --------------")
        while (self.forward_ready or self.back_ready or self._schedule_to_finish or not finished):
            # Conditional timeout context
            with contextlib.ExitStack() as stack:
                if self.step_timeout:
                    stack.enter_context(stopit.ThreadingTimeout(self.step_timeout, swallow_exc=False))

                finished = not bool(self.forward_ready or self.back_ready
                                    or self._schedule_to_finish)

                timestep += 1
                reg['sim/timestep'] = timestep
                if (timeout is not None) and (timestep == timeout):
                    break

                # if (timestep % 1000) == 0:
                #     log.info("-------------- Simulation cycle --------------")

                # print(f"-------------- {timestep} ------------------")
                # print(f'Tasks: {len(self.tasks)}')

                # if timestep == 516:
                #     breakpoint()

                self.phase = 'forward'
                self.sim_list(self.sim_gears)

                self.phase = 'delta'
                delta.set()
                delta.clear()

                self.phase = 'back'
                while self._schedule_to_finish:
                    for sim_gear in self._schedule_to_finish.copy():
                        self._finish(sim_gear)
                        self._schedule_to_finish.remove(sim_gear)

                self.sim_list(self.sim_gears)

                self.phase = 'cycle'
                self.events['before_timestep'](self, timestep)

                clk.set()
                clk.clear()

                # for sim_gear in reversed(self.sim_gears):
                #     if sim_gear in self.delta_ready:
                #         # if hasattr(sim_gear, 'port'):
                #         #     print(f'Clock: {sim_gear.gear.name}.{sim_gear.port.basename}')
                #         # else:
                #         #     print(f'Clock: {sim_gear.gear.name}')

                #         self.maybe_run_gear(sim_gear, self.delta_ready)

                for sim_gear in self.done:
                    self.remove(sim_gear)

                self.done.clear()

                self.events['after_timestep'](self, timestep)

        log.info(f"----------- Simulation done ---------------")
        log.info(f'Elapsed: {time.time() - start_time:.2f}')

        # while self.schedule_to_finish:
        #     # print(f'Canceling {sim_gear.gear.name}')
        #     sim_gear = self.schedule_to_finish.pop()
        #     self.finish(sim_gear)

    def cleanup(self):
        self.sim_map.clear()
        self.tasks.clear()
        self.task_data.clear()
        self.back_ready.clear()
        self.forward_ready.clear()
        self._schedule_to_finish.clear()
        self.sim_gears.clear()
        self.done.clear()
        self.wait_list.clear()
        self.cur_gear = None
        for e in self.events.values():
            e.clear()


    def run(self, timeout=None):
        self.sim_map = reg['sim/map']
        self.sim_gears = []
        self.tasks = {}
        self.task_data = {}

        v = GearEnum()
        v.visit(reg['gear/root'])

        self.insert_gears(simgear_exec_order(v.gears))

        self.wait_list = {}
        self.forward_ready = set(self.sim_gears)
        self.back_ready = set()
        self._schedule_to_finish = set()
        self.done = set()

        reg['sim/clk_event'] = asyncio.Event()
        reg['sim/delta_event'] = asyncio.Event()
        reg['sim/timestep'] = None

        self.events['before_setup'](self)

        for sim_gear in set(self.sim_gears):
            self.cur_gear = sim_gear.gear
            reg['gear/current_module'] = self.cur_gear
            reg['gear/current_sim'] = sim_gear
            sim_gear.setup()
            self.cur_gear = reg['gear/root']
            reg['gear/current_module'] = self.cur_gear
            reg['gear/current_sim'] = sim_gear

        reg['sim/exception'] = None
        reg['gear/exec_context'] = 'sim'
        try:
            self.events['before_run'](self)
            self.sim_loop(timeout)
        except SimFinish:
            pass
        except stopit.TimeoutException:
            reg['sim/traceback'] = sys.exc_info()[2]
            reg['sim/exception'] = SimTimeout(
                f'Timestep took longer then {self.step_timeout}s, simulation ended')
        except Exception as e:
            reg['sim/traceback'] = sys.exc_info()[2]
            reg['sim/exception'] = e

        if reg['sim/postmortem'] and reg['sim/traceback']:
            import pdb
            pdb.post_mortem(reg['sim/traceback'])

        reg['gear/exec_context'] = 'compile'

        try:
            # print(f"----------- After run ---------------")

            if not reg['sim/exception']:
                self.events['after_run'](self)
        except:
            pass

        for sim_gear in self.sim_gears:
            if not sim_gear.done:
                try:
                    sim_gear._finish()
                except:
                    pass

        try:
            self.events['after_cleanup'](self)
            self.events['at_exit'](self)
        finally:
            if reg['sim/exception']:
                raise reg['sim/exception']

            self.cleanup()


class SimSetupDone(Exception):
    pass


# TODO: This function throws an error when run two times in a script
def sim(resdir=None, timeout=None, extens=None, run=True, check_activity=False, seed=None):
    if reg['sim/dryrun']:
        raise SimSetupDone

    if extens is None:
        extens = []

    extens.extend(reg['sim/extens'])

    if resdir is None:
        resdir = reg['results-dir']

    reg['results-dir'] = resdir
    os.makedirs(resdir, exist_ok=True)

    if reg['sim/rand_seed'] is None:
        if seed is None:
            seed = random.randrange(sys.maxsize)

        reg['sim/rand_seed'] = seed

    random.seed(reg['sim/rand_seed'])

    log.info(f'Running sim with seed: {reg["sim/rand_seed"]}')

    loop = EventLoop(reg['sim/step_timeout'])
    asyncio.set_event_loop(loop)
    reg['sim/simulator'] = loop

    if check_activity:
        from pygears.sim.extens.activity import ActivityChecker
        if ActivityChecker not in extens:
            extens.append(ActivityChecker)

    for oper in itertools.chain(reg['sim/flow'], extens):
        oper()

    if run:
        loop.run(timeout)

        asyncio.new_event_loop()

        reg['sim/simulator'] = None
    else:
        return loop


class SimFmtFilter(LogFmtFilter):
    def filter(self, record):
        super().filter(record)

        m = reg['gear/current_module']

        record.module = m.name
        record.timestep = timestep()
        if record.timestep is None:
            record.timestep = '-'

        return True


class SimLog(CustomLogger):
    def get_format(self):
        return logging.Formatter(
            '%(timestep)s %(module)20s [%(levelname)s]: %(message)s %(err_file)s %(stack_file)s')

    def get_filter(self):
        return SimFmtFilter()


class SimPlugin(GearPlugin):
    @classmethod
    def bind(cls):
        global gear_reg, sim_reg
        gear_reg = {}
        sim_reg = {}

        reg['sim/config'] = {}
        reg['sim/flow'] = []
        reg['sim/tasks'] = {}
        reg['sim/simulator'] = None
        reg['sim/step_timeout'] = 0

        reg['sim/dryrun'] = False
        reg['sim'].subreg('hook')
        reg['sim/hook/cosim_build_before'] = SimEvent()
        reg['sim/hook/cosim_build_after'] = SimEvent()

        reg['gear/exec_context'] = None
        reg['sim/postmortem'] = False
        reg['sim/exception'] = None
        reg['sim/traceback'] = None
        reg.confdef('sim/rand_seed', None)
        reg.confdef('sim/clk_freq', 1000)
        reg.confdef('results-dir', default=tempfile.mkdtemp())
        reg.confdef('sim/extens', default=[])

        reg['gear/params/meta/sim_setup'] = None
        register_custom_log('sim', cls=SimLog)

        # temporary hack for pytest logger reset issue
        reg['logger/sim/error'] = 'exception'
        reg['logger/sim/print_traceback'] = False

    @classmethod
    def clear(cls):
        if reg['sim/simulator'] is not None:
            asyncio.new_event_loop()
            reg['sim/simulator'] = None

        reg['sim/tasks'] = {}


def sim_assert(cond, msg=None):
    if not cond:
        log.error(f'Assertion failed: {msg}')



import pydot
from pygears.core.port import InPort
from pygears.core.hier_node import HierVisitorBase
from pygears.util.find import find


class Visitor(HierVisitorBase):
    def __init__(self, node_filter):
        self.dot = GearGraph(graph_type='digraph', rankdir='TB', overlap=True)

        self.hier = [self.dot]
        self.node_filter = node_filter

    def enter_hier(self, module):
        self.hier.append(self.dot.cluster_map[module])

    def exit_hier(self, module):
        node = self.hier.pop()
        self.hier[-1].add_subgraph(node)

    def Gear(self, module):
        if self.node_filter(module):
            self.dot.node_map[module] = pydot.Node(
                module.name,
                fontsize=18,
                margin=0.01,
                tooltip=module.name,
                label=module.basename,
                shape="doubleoctagon")

            self.hier[-1].add_node(self.dot.node_map[module])
        else:
            self.dot.cluster_map[module] = pydot.Cluster(
                graph_name=module.name,
                label=module.basename,
                tooltip=module.name,
                fontsize=48,
                fontcolor='blue',
                labeljust='l',
                overlap=False)

            self.enter_hier(module)

            super().HierNode(module)

            self.exit_hier(module)

        return True


class GearGraph(pydot.Dot):
    def __init__(self, *argsl, **argsd):
        super().__init__(*argsl, **argsd)
        self._node_map = {}
        self._cluster_map = {}
        self._prod_edge_map = {}
        self._cons_edge_map = {}

    @property
    def node_map(self):
        return self._node_map

    @property
    def cluster_map(self):
        return self._cluster_map

    @property
    def prod_edge_map(self):
        return self._prod_edge_map

    @property
    def cons_edge_map(self):
        return self._cons_edge_map


def _get_consumer_tree_rec(intf, consumers, node_filter):
    for port in intf.consumers:
        cons_intf = port.consumer
        if node_filter(port.gear) and (isinstance(port, InPort)):
            consumers.append(port)
        else:
            _get_consumer_tree_rec(cons_intf, consumers, node_filter)


def get_consumer_tree(intf, node_filter):
    consumers = []
    _get_consumer_tree_rec(intf, consumers, node_filter)
    return consumers


def graph(top=None,
          node_filter=lambda x: not x.child,
          edge_fmt='{prod_gear} -> {cons_gear}'):
    if top is None:
        top = find('/')

    v = Visitor(node_filter)
    v.visit(top)

    edge_fmt_dict = {}

    dot = v.dot

    for module, node in dot.node_map.items():
        for pout in module.out_ports:
            edge_fmt_dict['prod_port'] = pout.basename
            edge_fmt_dict['prod_gear'] = pout.gear.basename

            if getattr(pout.consumer, 'var_name', None):
                edge_fmt_dict['prod_var'] = pout.consumer.var_name
            else:
                edge_fmt_dict['prod_var'] = edge_fmt_dict['prod_gear']

            edges = get_consumer_tree(pout.producer, node_filter)

            dot.prod_edge_map[pout] = []

            for e in edges:
                edge_fmt_dict['cons_port'] = e.basename
                edge_fmt_dict['cons_gear'] = e.gear.basename
                if getattr(e.producer, 'var_name', None):
                    edge_fmt_dict['cons_var'] = e.producer.var_name
                else:
                    edge_fmt_dict['cons_var'] = edge_fmt_dict['cons_gear']

                dot_edge = pydot.Edge(
                    node,
                    dot.node_map[e.gear],
                    fontsize=20,
                    decoreate=True,
                    label=eval(f'f"{edge_fmt}"', globals(), edge_fmt_dict))

                dot.cons_edge_map[e] = dot_edge
                dot.prod_edge_map[pout].append(dot_edge)
                dot.add_edge(dot_edge)

    return dot



import inspect

from . import ir
from pygears.typing import Bool, typeof
from pygears import Intf


def is_intf_id(expr):
    return isinstance(expr, ir.Name) and typeof(expr.dtype, ir.IntfType)
    # return (isinstance(expr, ir.Name) and isinstance(expr.obj, ir.Variable)
    #         and isinstance(expr.obj.val, Intf))


def add_to_list(orig_list, extension):
    if extension:
        orig_list.extend(extension if isinstance(extension, list) else [extension])


res_true = ir.ResExpr(Bool(True))
res_false = ir.ResExpr(Bool(False))


class Scope:
    def __init__(self, parent=None):
        self.parent = None
        self.child = None
        self.items = {}

    def subscope(self):
        if self.child is None:
            self.child = Scope(parent=self)
            self.child.parent = self
            return self.child

        return self.child.subscope()

    def upscope(self):
        s = self.cur_subscope
        s.parent.child = None
        s.parent = None

    @property
    def cur_subscope(self):
        if self.child is None:
            return self

        return self.child.cur_subscope

    @property
    def top_scope(self):
        if self.parent is not None:
            return self.parent.top_scope

        return self

    def clear(self):
        self.child.clear()
        self.child = None
        self.items.clear()

    def __getitem__(self, key):
        if self.child:
            try:
                return self.child[key]
            except KeyError:
                pass

        return self.items[key]

    def __delitem__(self, key):
        if self.child:
            try:
                del self.child[key]
            except KeyError:
                pass

        del self.items[key]

    def __contains__(self, key):
        if self.child and key in self.child:
            return True

        return key in self.items

    def __setitem__(self, key, val):
        if self.child:
            self.child[key] = val
            return

        self.items[key] = val

    def __iter__(self):
        scope = self.cur_subscope
        keys = set()
        while True:
            for key in scope.items:
                if key not in keys:
                    yield key
                    keys.add(key)

            if scope.parent is None:
                return

            scope = scope.parent


class HDLVisitor:
    def __init__(self, ctx):
        self.ctx = ctx

    def visit(self, node):
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                return getattr(self, base_class.__name__)(node)
        else:
            return self.generic_visit(node)

    def BaseBlock(self, block: ir.BaseBlock):
        for stmt in block.stmts:
            self.visit(stmt)

        return block

    def HDLBlock(self, block: ir.HDLBlock):
        for b in block.branches:
            self.visit(b)

    # def ExprStatement(self, stmt: ir.ExprStatement):
    #     self.visit(stmt.expr)
    #     return stmt

    # def AssignValue(self, stmt: ir.AssignValue):
    #     self.visit(stmt.val)
    #     self.visit(stmt.target)
    #     return stmt

    def generic_visit(self, node):
        return node


class HDLScopedVisitor(HDLVisitor):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.scopes = []

    @property
    def parent(self):
        return self.scopes[-1]

    def BaseBlock(self, block: ir.BaseBlock):
        self.scopes.append(block)
        super().BaseBlock(block)
        self.scopes.append(block)

    def HDLBlock(self, block: ir.HDLBlock):
        self.scopes.append(block)
        super().BaseBlock(block)
        self.scopes.append(block)

    def generic_visit(self, node):
        return node


class IrVisitor:
    def __init__(self):
        self.scopes = []

    @property
    def parent(self):
        return self.scopes[-1]

    def enter_scope(self, block):
        self.scopes.append(block)

    def exit_scope(self):
        self.scopes.pop()

    def visit(self, node):
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                getattr(self, base_class.__name__)(node)
                return
        else:
            self.generic_visit(node)

    def BaseBlock(self, block: ir.BaseBlock):
        self.enter_scope(block)
        for stmt in block.stmts:
            self.visit(stmt)
        self.exit_scope()

    def Branch(self, block: ir.Branch):
        self.visit(block.test)
        self.BaseBlock(block)

    def LoopBlock(self, block: ir.LoopBlock):
        self.BaseBlock(block)
        self.visit(block.test)

    def HDLBlock(self, block: ir.HDLBlock):
        self.enter_scope(block)
        for b in block.branches:
            self.visit(b)
        self.exit_scope()

    def FuncReturn(self, stmt: ir.FuncReturn):
        self.visit(stmt.expr)

    def ExprStatement(self, stmt: ir.ExprStatement):
        self.visit(stmt.expr)

    def AssignValue(self, stmt: ir.AssignValue):
        self.visit(stmt.target)
        self.visit(stmt.val)

    def AssertValue(self, stmt: ir.AssertValue):
        self.visit(stmt.val)

    def generic_visit(self, node):
        pass


class IrRewriter:
    def __init__(self):
        self.scopes = []

    def visit(self, node):
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                return getattr(self, base_class.__name__)(node)
        else:
            return self.generic_visit(node)

    def enter(self, node):
        for base_class in inspect.getmro(node.__class__):
            method_name = f'enter_{base_class.__name__}'
            if hasattr(self, method_name):
                return getattr(self, method_name)(node)

    def exit(self, node):
        for base_class in inspect.getmro(node.__class__):
            method_name = f'exit_{base_class.__name__}'
            if hasattr(self, method_name):
                return getattr(self, method_name)(node)

    @property
    def parent(self):
        return self.scopes[-1]

    def enter_scope(self, block):
        self.scopes.append(block)

    def exit_scope(self):
        self.scopes.pop()

    def BaseBlock(self, block: ir.BaseBlock):
        rw_block = type(block)()
        self.enter(rw_block)
        self.enter_scope(rw_block)
        for stmt in block.stmts:
            add_to_list(rw_block.stmts, self.visit(stmt))

        self.exit_scope()
        self.exit(rw_block)
        return rw_block

    def Branch(self, block: ir.Branch):
        rw_block = type(block)()
        rw_block.test = self.visit(block.test)
        self.enter(rw_block)
        self.enter_scope(rw_block)

        for stmt in block.stmts:
            add_to_list(rw_block.stmts, self.visit(stmt))

        self.exit_scope()
        self.exit(rw_block)
        return rw_block

    def LoopBlock(self, block: ir.LoopBlock):
        rw_block = type(block)(blocking=block.blocking)
        self.enter(rw_block)
        self.enter_scope(rw_block)

        for stmt in block.stmts:
            add_to_list(rw_block.stmts, self.visit(stmt))

        self.exit_scope()

        rw_block.test = self.visit(block.test)
        self.exit(rw_block)
        return rw_block

    def FuncBlock(self, block: ir.FuncBlock):
        # args = {n: self.visit(val) for n, val in block.args.items()}

        rw_block = type(block)(name=block.name,
                               args=block.args,
                               ret_dtype=block.ret_dtype,
                               funcs=block.funcs)

        self.enter(rw_block)
        self.enter_scope(rw_block)
        for stmt in block.stmts:
            add_to_list(rw_block.stmts, self.visit(stmt))

        self.exit_scope()
        self.exit(rw_block)
        return rw_block

    def HDLBlock(self, block: ir.HDLBlock):
        rw_block = type(block)()
        self.enter(rw_block)
        self.enter_scope(rw_block)
        branches = [self.visit(b) for b in block.branches]

        rw_block.branches = branches

        self.exit_scope()
        self.exit(rw_block)
        return rw_block

    def ExprStatement(self, stmt: ir.ExprStatement):
        return type(stmt)(self.visit(stmt.expr))

    def FuncReturn(self, stmt: ir.FuncReturn):
        return type(stmt)(stmt.func, self.visit(stmt.expr))

    def AssignValue(self, stmt: ir.AssignValue):
        return type(stmt)(self.visit(stmt.target), self.visit(stmt.val))

    def AssertValue(self, stmt: ir.AssertValue):
        return type(stmt)(self.visit(stmt.val))

    def Expr(self, expr):
        return expr

    def generic_visit(self, node):
        return node


class IrExprVisitor:
    def visit(self, node):
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_CallExpr(self, node: ir.CallExpr):
        self.visit(node.func)

        for arg in node.args:
            self.visit(arg)

        for name, arg in node.kwds.items():
            self.visit(arg)

    def visit_FunctionCall(self, node: ir.FunctionCall):
        for arg in node.operands:
            self.visit(arg)

        if node.keywords:
            for name, arg in node.keywords.items():
                self.visit(arg)

    def visit_AttrExpr(self, node):
        self.visit(node.val)

    def visit_CastExpr(self, node):
        self.visit(node.operand)

    def visit_ConcatExpr(self, node):
        for op in node.operands:
            self.visit(op)

    def visit_ArrayOpExpr(self, node):
        self.visit(node.array)

    def visit_UnaryOpExpr(self, node):
        self.visit(node.operand)

    def visit_BinOpExpr(self, node):
        for op in node.operands:
            self.visit(op)

    def visit_SubscriptExpr(self, node):
        self.visit(node.val)
        self.visit(node.index)

    def visit_ConditionalExpr(self, node):
        self.visit(node.cond)
        for op in node.operands:
            self.visit(op)

    def generic_visit(self, node):
        pass


# TODO: Consolidate method naming with IrVisitor
class IrExprRewriter:
    def visit(self, node):
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_AttrExpr(self, node):
        val = self.visit(node.val)
        if val is not None:
            return ir.AttrExpr(val, node.attr)

        return node

    def visit_CastExpr(self, node):
        operand = self.visit(node.operand)
        if operand is not None:
            return ir.CastExpr(operand, node.cast_to)

        return node

    def visit_CallExpr(self, node):
        args = [self.visit(a) for a in node.args]
        kwds = {name: self.visit(val) for name, val in node.kwds.items()}
        func = self.visit(node.func)

        return ir.CallExpr(func, args, kwds, node.params)

    def visit_FunctionCall(self, node: ir.FunctionCall):
        changed = False
        args = []
        for arg in node.operands:
            res = self.visit(arg)
            if res is None:
                args.append(arg)
            else:
                changed = True
                args.append(res)

        kwds = {}
        if node.keywords:
            for name, arg in node.keywords.items():
                res = self.visit(arg)
                if res is None:
                    kwds[name] = arg
                else:
                    changed = True
                    kwds[name] = res

        if not changed:
            return node

        return ir.FunctionCall(node.name, args, kwds, node.ret_dtype)

    def visit_ConcatExpr(self, node):
        ops = [self.visit(op) for op in node.operands]
        if all(op is None for op in ops):
            return node

        ops = [old_op if new_op is None else new_op for new_op, old_op in zip(ops, node.operands)]

        return ir.ConcatExpr(tuple(ops))

    def visit_ArrayOpExpr(self, node):
        array = self.visit(node.array)
        if array is not None:
            return ir.ArrayOpExpr(array, node.operator)

        return node

    def visit_UnaryOpExpr(self, node):
        operand = self.visit(node.operand)
        if operand is not None:
            return ir.UnaryOpExpr(operand, node.operator)

        return node

    def visit_BinOpExpr(self, node):
        ops = [self.visit(op) for op in node.operands]
        if all(op is None for op in ops):
            return node

        ops = [old_op if new_op is None else new_op for new_op, old_op in zip(ops, node.operands)]

        return ir.BinOpExpr(tuple(ops), node.operator)

    def visit_SubscriptExpr(self, node):
        old_ops = (node.val, node.index)

        ops = [self.visit(op) for op in old_ops]
        if all(op is None for op in ops):
            return node

        ops = [old_op if new_op is None else new_op for new_op, old_op in zip(ops, old_ops)]

        return ir.SubscriptExpr(*ops, ctx=node.ctx)

    def visit_ConditionalExpr(self, node):
        old_ops = (node.cond, *node.operands)

        ops = [self.visit(op) for op in old_ops]
        if all(op is None for op in ops):
            return node

        ops = [old_op if new_op is None else new_op for new_op, old_op in zip(ops, old_ops)]

        return ir.ConditionalExpr(tuple(ops[1:]), ops[0])

    def generic_visit(self, node):
        return node


__all__ = ['Scope', 'HDLVisitor', 'res_true', 'res_false', 'add_to_list', 'ir']



import json
from pygears import PluginBase, find, reg
from pygears.core.port import OutPort
from pygears.sim import timestep
from pygears.typing import typeof, TLM, Queue
from .vcd import VCDTypeVisitor, VCDValVisitor, is_trace_included
from pygears.core.hier_node import HierVisitorBase
from .sim_extend import SimExtend
import os
import itertools
import atexit
from pygears.conf import inject, Inject
from dataclasses import dataclass
from typing import Any


@dataclass
class WaveIntf():
    waves: dict
    dtype: Any = None


def register_traces_for_intf(dtype, scope, writer):
    vcd_vars = {}

    if typeof(dtype, TLM):
        vcd_vars['data'] = writer.register_var(scope, 'data', 'string')
    else:
        v = VCDTypeVisitor()
        v.visit(dtype, '')

        for name, t in v.fields.items():
            if name:
                field_scope, _, basename = name.rpartition('.')
                if field_scope:
                    field_scope = [scope] + field_scope.split('.')
                else:
                    field_scope = [scope]
            else:
                basename = ''
                field_scope = [scope]

            vcd_vars[name] = writer.register(basename, field_scope)

    return WaveIntf(vcd_vars, dtype=dtype)


class WaveJSONHierVisitor(HierVisitorBase):
    @inject
    def __init__(self, writer, include, sim_map=Inject('sim/map')):

        self.include = include
        self.vcd_tlm = False
        self.sim_map = sim_map
        self.vcd_vars = {}
        self.writer = writer
        self.indent = 0

    def enter_hier(self, name):
        self.indent += 4

    def exit_hier(self, name):
        self.indent -= 4

    def Gear(self, module):
        self.enter_hier(module.basename)

        # if module in self.sim_map and module.params['sim_cls'] is None:
        gear_vcd_scope = module.name[1:].replace('/', '.')
        for p in itertools.chain(module.in_ports, module.out_ports):
            if not is_trace_included(p, self.include, self.vcd_tlm):
                continue

            scope = '.'.join([gear_vcd_scope, p.basename])
            if isinstance(p, OutPort):
                intf = p.producer
            else:
                intf = p.consumer

            try:
                in_queue = intf.in_queue
            except:
                continue

            # TODO: support list of in_queues properly, which means that this
            # port is in front of a broadcast.
            if isinstance(in_queue, list):
                in_queue = in_queue[0]

            if in_queue:
                intf = in_queue.intf

            intf_vars = self.vcd_vars.get(intf, [])

            intf_vars.append(
                register_traces_for_intf(p.dtype, scope, self.writer))

            self.vcd_vars[intf] = intf_vars

        super().HierNode(module)

        self.exit_hier(module.basename)

        return True


@dataclass(frozen=True)
class WaveSignal:
    name: str
    scope: tuple


@dataclass
class WaveValue():
    __slots__ = ['state', 'val']
    state: int
    val: Any


class WaveJSONWriter:
    def __init__(self):
        self.signals = []
        self.values = {}

    def register(self, name, scope=None):
        sig = WaveSignal(name, tuple(scope) if scope is not None else tuple())
        self.signals.append(sig)
        self.values[sig] = {}
        return sig

    def change(self, sig, value, timestep, state=1):
        self.values[sig][timestep] = WaveValue(state, value)

    def ack(self, sig, timestep):
        vals = self.values[sig]
        if timestep in vals:
            self.values[sig][timestep].state = 3
        else:
            self.values[sig][timestep] = WaveValue(3, None)

    def json(self, vcd_vars):
        if timestep() is None:
            return {}

        def get_sig_scope(data, scope):
            if not scope:
                return data

            for child in data:
                if not isinstance(child, list):
                    continue

                if child[0] == scope[0]:
                    return get_sig_scope(child, scope[1:])

            data.append([scope[0]])
            return get_sig_scope(data[-1], scope[1:])

        data = {
            'signal': [{
                'name': 'clk',
                'wave': 'p' + '.' * (timestep())
            }],
            'head': {
                'tock': 0
            },
        }
        for _, wintfs in vcd_vars.items():
            for w in wintfs:
                for name, s in w.waves.items():

                    # for s in self.signals:
                    vals = self.values[s]

                    eot = 0
                    js_sig = {'name': s.name, 'wave': '', 'data': []}
                    state = None
                    for t in range(timestep() + 1):
                        if t in vals:
                            if typeof(w.dtype, Queue):
                                if self.values[w.waves['eot']][t].val:
                                    eot = 1
                                elif self.values[w.waves['eot']][t].val == 0:
                                    eot = 0

                            val = vals[t]
                            if val.state == 1:
                                js_sig['wave'] += '4'
                            elif eot:
                                js_sig['wave'] += '3'
                            else:
                                js_sig['wave'] += '5'

                            state = val.state
                            if val.val is not None:
                                js_sig['data'].append(str(val.val))
                            else:
                                js_sig['data'].append(js_sig['data'][-1])

                        elif state == 3:
                            state = 0
                            js_sig['wave'] += 'z'
                        elif state is None:
                            state = 0
                            js_sig['wave'] += 'z'
                        else:
                            js_sig['wave'] += '.'

                    scope = get_sig_scope(data['signal'], s.scope)
                    scope.append(js_sig)

        return data


class WaveJSONValVisitor(VCDValVisitor):
    def change(self, dtype, field, val):
        self.writer.change(self.vcd_vars[field], dtype(val).width,
                           self.timestep)


class WaveJSON(SimExtend):
    @inject
    def __init__(self,
                 trace_fn=Inject('wavejson/trace_fn'),
                 include=Inject('debug/trace'),
                 sim=Inject('sim/simulator'),
                 outdir=Inject('results-dir'),
                 sim_map=Inject('sim/map')):
        super().__init__()
        self.sim = sim
        self.finished = False

        self.outdir = outdir
        self.trace_fn = trace_fn

        if not os.path.isabs(self.trace_fn):
            self.trace_fn = os.path.abspath(os.path.join(
                self.outdir, trace_fn))

        atexit.register(self.finish)

        self.writer = WaveJSONWriter()

        reg['WaveJSONWriter'] = self.writer

        self.handhake = set()

        v = WaveJSONHierVisitor(self.writer, include)
        v.visit(find('/'))
        self.vcd_vars = v.vcd_vars

        for intf in self.vcd_vars:
            intf.events['put'].append(self.intf_put)
            intf.events['ack'].append(self.intf_ack)

    def intf_put(self, intf, val):
        if intf not in self.vcd_vars:
            return True

        for v in self.vcd_vars[intf]:
            if typeof(intf.dtype, TLM):
                self.writer.change(v.waves['data'], timestep(), str(val))
            else:
                visitor = WaveJSONValVisitor(v.waves, self.writer, timestep())
                visitor.visit(v.dtype, '', val=val)

        return True

    def intf_ack(self, intf):
        if intf not in self.vcd_vars:
            return True

        ts = timestep()
        for v in self.vcd_vars[intf]:
            for sig in v.waves.values():
                self.writer.ack(sig, ts)

        self.handhake.add(intf)
        return True

    def finish(self):
        if not self.finished:
            with open(self.trace_fn, 'w') as f:
                data = self.writer.json(self.vcd_vars)
                json.dump(data, f)

            self.finished = True

    def after_cleanup(self, sim):
        self.finish()


class WaveJSONPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg.confdef('wavejson/trace_fn', 'pygears.json')



import os
from pygears import reg


def _find_rec(path, root):
    parts = path.split("/")

    if parts[0] == '..':
        return _find_rec("/".join(parts[1:]), root.parent)

    for c in root.child:
        if hasattr(c, 'basename') and c.basename == parts[0]:
            child = c
            break
    else:
        raise ModuleNotFoundError()

    if len(parts) == 1:
        return child
    else:
        return _find_rec("/".join(parts[1:]), child)


def find(path, root=None):
    if path.startswith('/'):
        path = path[1:]
        root = reg['gear/root']
    else:
        if path.startswith('./'):
            path = path[2:]

        root = reg['gear/current_module']

    if path == '':
        return root

    module_path, port_name = os.path.splitext(path)

    try:
        module = _find_rec(module_path, root)
    except ModuleNotFoundError:
        return None

    if not port_name:
        return module
    else:
        port_name = port_name[1:]
        for p in module.in_ports + module.out_ports + module.local_intfs:
            if p.basename == port_name:
                return p

        return None



from pygears import alternative, gear
from pygears.typing import Tuple, Uint, Any, Queue, bitw
from pygears.util.utils import qrange
# from pygears.lib.rng import qrange


@gear
async def serialize(din: Tuple[{
        'data': Any,
        'active': Uint
}]) -> Queue['din["data"][0]']:
    async with din as (data, active):
        for i, last in qrange(active):
            yield data[i], last


@alternative(serialize)
@gear
async def serialize_plain(din) -> Queue['din[0]']:
    i = Uint[bitw(len(din.dtype) - 1)](0)

    async with din as val:
        for i, last in qrange(len(din.dtype)):
            yield val[i], last



import os

from pygears.conf import PluginBase, reg
from .inst import svgen_inst
from .svmod import SVModuleInst
from .resolvers import HDLFileResolver, HDLTemplateResolver, HierarchicalResolver, HLSResolver, BlackBoxResolver
from pygears.conf import PluginBase, reg

from pygears.hdl.templenv import TemplateEnv
from .v.util import vgen_intf, vgen_signal
from .util import svgen_typedef


class SVTemplateEnv(TemplateEnv):
    lang = 'sv'

    def __init__(self):
        super().__init__(basedir=os.path.dirname(__file__))

        self.jenv.globals.update(svgen_typedef=svgen_typedef)

        self.snippets = self.load(self.basedir, 'snippet.j2').module


class SVGenPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['svgen/flow'] = [svgen_inst]
        reg['svgen/resolvers'] = [
            HDLFileResolver, HDLTemplateResolver, HierarchicalResolver, HLSResolver
        ]
        reg['svgen/dflt_resolver'] = BlackBoxResolver
        reg['svgen/templenv'] = SVTemplateEnv()

        reg['svgen/module_namespace'] = {'Gear': SVModuleInst, 'GearHierRoot': SVModuleInst}


class VTemplateEnv(TemplateEnv):
    lang = 'v'

    def __init__(self):
        super().__init__(basedir=os.path.join(os.path.dirname(__file__), 'v'))

        self.jenv.globals.update(vgen_intf=vgen_intf, vgen_signal=vgen_signal)

        self.snippets = self.load(self.basedir, 'snippet.j2').module


class VGenPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['vgen/templenv'] = VTemplateEnv()

        reg['vgen/module_namespace'] = {'Gear': SVModuleInst, 'GearHierRoot': SVModuleInst}

        reg['vgen/flow'] = [svgen_inst]
        reg['vgen/resolvers'] = [
            HDLFileResolver,
            HDLTemplateResolver,
            HierarchicalResolver,
            HLSResolver,
        ]
        reg['vgen/dflt_resolver'] = BlackBoxResolver


from pygears.conf import load_plugin_folder
import os

load_plugin_folder(os.path.join(os.path.dirname(__file__), 'modules'))

__all__ = ['svgen_generate']



from functools import singledispatch
from pygears.typing import typeof, Int, Fixp, Uint, Ufixp, is_type, Integral, Fixpnumber, Integer, code
from pygears.typing.uint import IntegerType, IntType, UintType
from pygears.typing.fixp import FixpnumberType, FixpType, UfixpType
from .cast import cast


@singledispatch
def type_trunc(cast_type, val):
    raise TypeError(
        f"Type '{repr(cast_type)}' unsupported, cannot truncate type '{val}' "
        f"of type '{repr(type(val))}'")


@type_trunc.register(IntegerType)
def integer_type_trunc_resolver(cast_type: IntegerType, dtype):
    if dtype is int:
        return cast_type

    if not is_type(dtype) or dtype.base != cast_type.base:
        raise TypeError(
            f"cannot truncate type '{repr(dtype)}' to a type '{repr(cast_type)}'"
            f" of a different base type")

    return cast_type


@type_trunc.register(FixpnumberType)
def fixp_type_trunc_resolver(cast_type: FixpnumberType, dtype):
    if not is_type(dtype) or dtype.base != cast_type.base:
        raise TypeError(
            f"cannot truncate type '{repr(dtype)}' to a type '{repr(cast_type)}'"
            f" of a different base type")

    return cast_type


@singledispatch
def value_trunc(cast_type, val):
    raise ValueError(
        f"Truncating to type '{repr(cast_type)}' unsupported, cannot truncate value '{val}' "
        f"of type '{repr(type(val))}'")


@value_trunc.register(FixpType)
def fixp_value_trunc_resolver(trunc_type: FixpType, val):
    bv = code(val, Uint)
    sign = bv >> (type(val).width - 1)

    if type(val).fract >= trunc_type.fract:
        bv_fract_trunc = bv >> (type(val).fract - trunc_type.fract)

    else:
        bv_fract_trunc = bv << (trunc_type.fract - type(val).fract)

    if type(val).integer >= trunc_type.integer:
        bv_res = (
            bv_fract_trunc & ((1 << (trunc_type.width - 1)) - 1) |
            (sign << (trunc_type.width - 1)))
    else:
        sign_exten = Uint[trunc_type.integer - type(val).integer].max if sign else 0
        bv_res = (sign_exten << (type(val).integer + trunc_type.fract)) | bv_fract_trunc

    return trunc_type.decode(bv_res)


@value_trunc.register(UfixpType)
def ufixp_value_trunc_resolver(trunc_type: UfixpType, val):
    if type(val).fract > trunc_type.fract:
        return trunc_type.decode(code(val, int) >> (type(val).fract - trunc_type.fract))
    else:
        return trunc_type.decode(code(val, int) << (trunc_type.fract - type(val).fract))


@value_trunc.register(IntType)
def int_value_trunc_resolver(trunc_type: IntType, val):
    bv = code(val, int)
    sign = bv >> (type(val).width - 1)

    if trunc_type.width <= type(val).width:
        bv_res = bv & ((1 << trunc_type.width) - 1) \
            | (sign << (trunc_type.width - 1))
    else:
        sign_exten = Uint[trunc_type.width - type(val).width].max if sign else 0
        bv_res = (sign_exten << type(val).width) | bv

    return trunc_type.decode(bv_res)


@value_trunc.register(UintType)
def uint_value_trunc_resolver(trunc_type: UintType, val):
    return trunc_type.decode(code(val, int) & ((1 << trunc_type.width) - 1))


def trunc(data, t):
    if is_type(data):
        return type_trunc(t, data)
    else:
        return value_trunc(type_trunc(t, type(data)), data)



import fnmatch
import functools
import typing
import os
import pygears
from pygears import reg
from pygears.core.gear import OutSig
from ...base_resolver import ResolverBase, ResolverTypeError
from pygears.util.fileio import find_in_dirs, save_file
from pygears.conf import inject, Inject
from pygears.hdl import hdlmod


class HierarchicalResolver(ResolverBase):
    @inject
    def __init__(self, node):
        self.node = node

        if not node.meta_kwds.get('hdl', {}).get('hierarchical', node.hierarchical):
            raise ResolverTypeError

    @property
    def hdl_path_list(self):
        return reg[f'{self.lang}gen/include']

    @property
    def files(self):
        files = [self.file_basename]
        if 'hdl' in self.node.meta_kwds:
            if 'files' in self.node.meta_kwds['hdl']:
                for fn in self.node.meta_kwds['hdl']['files']:
                    if not os.path.splitext(fn)[-1]:
                        fn = f'{fn}.{self.lang}'

                    files.append(fn)

        return files

    @property
    @functools.lru_cache()
    def module_name(self):
        if find_in_dirs(f'{self.hier_path_name}.{self.lang}',
                        self.hdl_path_list):
            return self.hier_path_name + '_hier'
        else:
            return self.hier_path_name

    @property
    def file_basename(self):
        return f'{self.module_name}.{self.lang}'

    def module_context(self, template_env):
        attrib = self.cfg.get('attrib', None)
        if isinstance(attrib, str):
            attrib = [attrib]

        context = {
            'pygears': pygears,
            'module_name': self.module_name,
            'intfs': template_env.port_intfs(self.node),
            # 'sigs': [s.name for s in self.node.meta_kwds['signals']],
            'sigs': self.node.meta_kwds['signals'],
            'params': self.node.params,
            'inst': [],
            'generics': [],
            'comment': {
                'comment': '',
                'attrib': [] if attrib is None else attrib
            }
        }

        for port in context['intfs']:
            context[f'_{port["name"]}'] = port
            context[f'_{port["name"]}_t'] = port['type']

        return context

    @property
    def params(self):
        return {}

    def get_hier_module(self, template_env):
        context = self.module_context(template_env)

        for child in self.node.local_intfs:
            hmod = hdlmod(child)
            contents = hmod.get_inst(template_env)
            if contents:
                context['inst'].append(contents)

        for child in self.node.child:
            for s in child.meta_kwds['signals']:
                if isinstance(s, OutSig):
                    name = child.params['sigmap'][s.name]
                    context['inst'].append(f'logic [{s.width-1}:0] {name};')

            hmod = hdlmod(child)
            if hasattr(hmod, 'get_inst'):
                contents = hmod.get_inst(template_env)
                if contents:
                    if hmod.traced:
                        context['inst'].append('/*verilator tracing_on*/')
                    context['inst'].append(contents)
                    if hmod.traced:
                        context['inst'].append('/*verilator tracing_off*/')

        return template_env.render_local(__file__, "hier_module.j2", context)

    def generate(self, template_env, outdir):
        save_file(self.file_basename, outdir,
                  self.get_hier_module(template_env))



from pygears import gear, alternative
from pygears.typing import Queue, Uint, Union, bitw, trunc
from .rng import qrange
from .flatten import flatten
from .ccat import ccat
from .demux import demux


@gear(enablement=b'lvl < din.lvl')
async def qdeal_impl(din: Queue, *, num,
                     lvl=b'din.lvl-1') -> b'Union[(Queue[din.data, din.lvl-1], ) * num]':

    i = Uint[bitw(num - 1)](0)

    async for (data, eot) in din:
        out_eot = eot[:lvl]

        d = data if lvl == 0 else (data, out_eot)

        yield (d, i)

        if all(out_eot):
            if i == (num - 1):
                i = 0
            else:
                i += 1


@alternative(qdeal_impl)
@gear(enablement=b'lvl == din.lvl')
async def qdeal_impl_same_lvl(din: Queue, *, num, lvl=b'din.lvl-1') -> b'Union[(din, ) * num]':
    for i in range(num):
        async for (data, eot) in din:
            d = data if lvl == 0 else (data, eot)

            yield (d, trunc(i, Uint[bitw(num - 1)]))


@gear
def qdeal(din: Queue, *, num, lvl=b'din.lvl-1'):
    return din \
        | qdeal_impl(num=num, lvl=lvl) \
        | demux(use_dflt=False,
                mapping={n: n
                         for n in range(num)})


@gear
def deal(din, *, num) -> b'(din, ) * num':
    return ccat(din, qrange(num - 1, inclusive=True) | flatten) \
        | Union \
        | demux(use_dflt=False, mapping={i: i for i in range(num)})



from pygears import gear
from pygears.typing import cast as type_cast
from pygears.typing import trunc as type_trunc
from pygears.conf import reg
from pygears.core.intf import IntfOperPlugin
from pygears.hdl.util import HDLGearHierVisitor, flow_visitor
from pygears.hdl.sv import SVGenPlugin
from pygears.hdl.v import VGenPlugin


@gear
async def trunc(din, *, t) -> b'type_trunc(din, t)':
    async with din as d:
        yield type_trunc(d, t)


@gear
async def cast(din, *, t) -> b'type_cast(din, t)':
    async with din as d:
        yield type_cast(d, t)


def pipe(self, other):
    if self.producer is not None:
        name = f'cast_{self.producer.basename}'
    else:
        name = 'cast'

    return cast(self, t=other, name=name)


@flow_visitor
class RemoveEqualReprCastVisitor(HDLGearHierVisitor):
    def cast(self, node):
        pout = node.out_ports[0]
        pin = node.in_ports[0]

        if pin.dtype.width == pout.dtype.width:
            node.bypass()


class HDLCastPlugin(IntfOperPlugin, VGenPlugin, SVGenPlugin):
    @classmethod
    def bind(cls):
        reg['gear/intf_oper/__or__'] = pipe
        reg['vgen/flow'].insert(0, RemoveEqualReprCastVisitor)
        reg['svgen/flow'].insert(0, RemoveEqualReprCastVisitor)



from pygears import module, datagear
from pygears.typing import Uint
from pygears.typing import Integral, code, saturate as type_saturate


@datagear
def saturate(data: Integral, *, t, limits=None) -> b'type_saturate(data, t, limits)':
    return type_saturate(data, t, limits)



import os
import copy
import inspect

from typing import List
from dataclasses import dataclass
from pygears.conf import PluginBase, reg
from pygears.core.graph import has_async_producer
from traceback import walk_stack
from .intf import Intf
from .port import InPort, OutPort, HDLConsumer, HDLProducer
from .hier_node import NamedHierNode, HierVisitorBase
from .util import is_standard_func


class TooManyArguments(Exception):
    pass


class GearTypeNotSpecified(Exception):
    pass


class GearArgsNotSpecified(Exception):
    pass


def module():
    return reg['gear/current_module']


def filter_internals(t):
    is_internal = t[0].f_code.co_filename.startswith(os.path.dirname(__file__))
    is_boltons = 'decorator-gen' in t[0].f_code.co_filename
    return not is_internal and not is_boltons


def enum_stacktrace():
    return filter(filter_internals, walk_stack(f=None))


class GearHierRoot(NamedHierNode):
    def __init__(self, name=''):
        super().__init__(name)
        self.in_ports = []
        self.out_ports = []
        self.params = {}
        self.params.update(copy.deepcopy(reg['gear/params/extra']))
        self.func = None
        self.const_args = []
        self.args = []
        self.trace = None

        def __main():
            pass

        self.definition = __main

    @property
    def hierarchical(self):
        return True


def gear_explicit_params(func, params):
    paramspec = inspect.getfullargspec(func)
    explicit_param_names = paramspec.kwonlyargs or []

    explicit_params = {name: params[name] for name in explicit_param_names if name in params}

    if paramspec.varkw:
        for k, v in params.items():
            if ((k not in reg['gear/params']['meta']) and (k not in reg['gear/params']['extra'])
                    and (k not in paramspec.args) and (k != 'return')):
                explicit_params[k] = v

    return explicit_params


def struct_copy(s):
    if isinstance(s, dict):
        return type(s)({k: struct_copy(v) for k, v in s.items()})

    if isinstance(s, list):
        return type(s)([struct_copy(v) for v in s])

    return s


class Gear(NamedHierNode):
    def __init__(self, func, params):
        super().__init__(params['name'], reg['gear/current_module'] if func else None)
        self.meta_kwds = getattr(func, 'meta_kwds', {}).copy()

        self.trace = list(enum_stacktrace())
        self.args = {}
        # self.params = struct_copy(params)

        self.params = params

        sigmap = self.params.get('sigmap', {})
        self.params['sigmap'] = self.meta_kwds.get('sigmap', {})
        self.params['sigmap'].update(sigmap)

        if '__outnames__' not in params:
            params['__outnames__'] = copy.copy(self.meta_kwds.get('outnames', []))

        for p in reg['gear/params/extra']:
            if p not in params:
                continue

            self.params[p] = struct_copy(params[p])

        self.func = func
        self.const_args = {}
        self.in_ports: List[InPort] = []
        self.out_ports: List[OutPort] = []

    def _clean():
        pass

    def __repr__(self):
        if self.parent is None:
            return f'Top'

        return f'{self.definition.__name__}("{self.name}")'

    def connect_input(self, args, const_args):
        for name, val in const_args.items():
            from pygears.lib import const
            const(val=val, intfs=[args[name]])

        for i, (name, intf) in enumerate(args.items()):
            port = InPort(self, i, name, dtype=self.params[name])
            intf.connect(port)
            self.in_ports.append(port)

        self.const_args = const_args
        self.args = args

    def connect_output(self, out_intfs, out_dtypes):

        dflt_dout_name = reg['gear/naming/default_out_name']
        for i in range(len(self.outnames), len(out_dtypes)):
            if out_intfs and hasattr(out_intfs[i], 'var_name'):
                self.outnames.append(out_intfs[i].var_name)
            else:
                self.outnames.append(dflt_dout_name if len(out_dtypes) ==
                                     1 else f'{dflt_dout_name}{i}')

        self.out_ports = [OutPort(self, i, name) for i, name in enumerate(self.outnames)]

        # Connect internal interfaces
        if out_intfs:
            for i, r in enumerate(out_intfs):
                r.connect(self.out_ports[i])
        else:
            for dtype, port in zip(out_dtypes, self.out_ports):
                Intf(dtype).connect(port)

        for name, dtype in zip(self.outnames, out_dtypes):
            self.params[name] = dtype

    @property
    def hierarchical(self):
        if self.out_ports:
            return not any(has_async_producer(p) for p in self.out_ports)
        else:
            for p in self.in_ports:
                if p.consumer is None:
                    return False

                if isinstance(p.consumer.consumers[0], HDLConsumer):
                    return False
            else:
                return True

    @property
    def definition(self):
        return self.meta_kwds.get('definition', None)

    @property
    def outnames(self):
        return self.params.get('__outnames__', None)

    @property
    def tout(self):
        if len(self.out_ports) > 1:
            return tuple(i.dtype for i in self.out_ports)
        elif len(self.out_ports) == 1:
            return self.out_ports[0].dtype
        else:
            return None

    @property
    def signals(self):
        return {sig.name: sig for sig in self.meta_kwds['signals']}

    @property
    def dout(self):
        ret = self.out_port_intfs
        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    @property
    def local_intfs(self):
        intfs = {}  # acts as ordered set
        for c in self.child:
            for i in c.inputs + c.outputs:
                intfs[i] = None

        return list(intfs.keys())

    @property
    def in_port_intfs(self):
        in_intfs = []
        for port in self.in_ports:
            if port.consumer is None:
                i = Intf(self.params[port.basename])
                i.source(port)

            in_intfs.append(port.consumer)

        return tuple(in_intfs)
        # return tuple(p.consumer for p in self.in_ports)

    @property
    def out_port_intfs(self):
        return tuple(p.producer for p in self.out_ports)

    @property
    def inputs(self):
        return tuple(p.producer for p in self.in_ports)

    @property
    def outputs(self):
        return tuple(p.consumer for p in self.out_ports)

    @property
    def explicit_params(self):
        return gear_explicit_params(self.func, self.params)

    def remove(self):
        for p in self.in_ports:
            if p.producer is not None:
                try:
                    p.producer.disconnect(p)
                except ValueError:
                    pass

        for p in getattr(self, 'out_ports', []):
            if p.producer is not None:
                p.producer.disconnect(p)

        try:
            super().remove()
        except ValueError:
            pass


class create_hier:
    def __init__(self, gear):
        self.gear = gear

    def __enter__(self):
        reg['gear/current_module'] = self.gear
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        reg['gear/current_module'] = self.gear.parent


@dataclass(frozen=True, eq=True)
class InSig:
    name: str
    width: int
    modport: str = 'input'


@dataclass(frozen=True, eq=True)
class OutSig:
    name: str
    width: int
    modport: str = 'output'

    def write(self, val):
        pass


class GearPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['gear/naming'] = {'default_out_name': 'dout'}

        reg['gear'].subreg('params')

        reg['gear/params'].subreg(
            'meta', {
                'enablement': True,
                'outnames': None,
                'signals': (InSig('clk', 1), InSig('rst', 1)),
                '__base__': None,
                '__outnames__': None
            })

        reg['gear/params'].subreg('extra', {'name': None, 'intfs': [], 'sigmap': {}})

        reg['gear/root'] = Gear(None, params={'name': ''})
        reg['gear/current_module'] = reg['gear/root']
        reg['gear/exec_context'] = 'compile'

    @classmethod
    def clear(cls):
        class GearCleaner(HierVisitorBase):
            def HierNode(self, node):
                if node.definition:
                    node.definition._cache.clear()

                super().HierNode(node)
                node.__dict__.clear()

        GearCleaner().visit(reg['gear/root'])
        reg['gear/root'] = Gear(None, params={'name': ''})



import os
import shutil
from pygears import reg, find
from pygears.core.hier_node import HierYielderBase
from pygears.core.gear import Gear
from pygears.definitions import LIB_VLIB_DIR, LIB_SVLIB_DIR
from pygears.util.fileio import find_in_dirs


class NodeYielder(HierYielderBase):
    def __init__(self, filt):
        self.filt = filt
        self.vgen_map = reg[f'hdlgen/map']

    def Gear(self, node):
        if node in self.vgen_map:
            vinst = self.vgen_map[node]

            if self.filt and not self.filt(vinst):
                return

        for intf in node.local_intfs:
            yield intf

        yield node

def node_hdl_files(node, outdir):
    for f in node.files:
        path = find_in_dirs(f, reg[f'{node.lang}gen/include'])
        if path is not None:
            yield path
        else:
            yield os.path.join(outdir, f)


def enum_hdl_files(top, outdir, rtl_only=False, wrapper=False, filt=None):
    if isinstance(top, str):
        top = find(top)

    vgen_map = reg[f'hdlgen/map']
    dirs = {}
    for lang in ['v', 'sv', 'vhd']:
        dirs[lang] = reg[f'{lang}gen/include']

    dti_yielded = False

    if reg['hdl/toplang'] == 'sv':
        dti_yielded = True
        yield os.path.join(LIB_SVLIB_DIR, 'dti.sv')

    for node in NodeYielder(filt).visit(top):
        if node not in vgen_map:
            continue

        vinst = vgen_map[node]

        if vinst.lang == 'sv' and not dti_yielded:
            dti_yielded = True
            yield os.path.join(LIB_SVLIB_DIR, 'dti.sv')

        # TODO: What?
        if ((node is top) and wrapper and not rtl_only):
            yield os.path.join(outdir, f'wrap_{vinst.file_basename}')

        if isinstance(node, Gear):
            for f in vinst.files:
                path = find_in_dirs(f, dirs[vinst.lang])
                if path is not None:
                    yield path
                else:
                    yield os.path.join(outdir, f)

        if hasattr(vinst, 'file_basename'):
            file_name = getattr(vinst, 'impl_path', None)
            if file_name:
                yield file_name
            # elif rtl_only is False:
            #     for f in vinst.files:
            #         yield os.path.join(outdir, f)

        elif vinst.is_broadcast:
            if vinst.lang == 'v':
                yield os.path.join(LIB_VLIB_DIR, 'bc.v')
            elif vinst.lang == 'sv':
                yield os.path.join(LIB_SVLIB_DIR, 'bc.sv')


def list_hdl_files(top, outdir, rtl_only=False, wrapper=False, filt=None):

    if isinstance(top, str):
        top = find(top)

    orig_fns = set()
    hdlmods = {}
    for fn in enum_hdl_files(top, outdir, rtl_only=rtl_only, wrapper=wrapper, filt=filt):
        modname, lang = os.path.splitext(os.path.basename(fn))
        hdlmods[(modname, lang[1:])] = fn

    hdltop = reg['hdlgen/map'][top]
    disambig = reg['hdlgen/disambig']
    fns = []
    for (modname, lang), fn in hdlmods.items():
        if lang == hdltop.lang or (modname, hdltop.lang) not in hdlmods:
            fns.append(fn)
            continue

        fn_dis = os.path.join(outdir, f'{modname}_{lang}.{lang}')
        disambig[(modname, lang)] = (fn, fn_dis)

        fns.append(fn_dis)

    return fns

    # seen = set()
    # seen_add = seen.add
    # return [
    #     x for x in enum_hdl_files(
    #         top, outdir, rtl_only=rtl_only, wrapper=wrapper)
    #     if not (x in seen or seen_add(x))
    # ]



import logging


def debug(msg, *args, **kwds):
    return logging.getLogger('sim').debug(msg, *args, **kwds)


def info(msg, *args, **kwds):
    return logging.getLogger('sim').info(msg, *args, **kwds)


def warning(msg, *args, **kwds):
    return logging.getLogger('sim').warning(msg, *args, **kwds)


def error(msg, *args, **kwds):
    return logging.getLogger('sim').error(msg, *args, **kwds)


def critical(msg, *args, **kwds):
    return logging.getLogger('sim').critical(msg, *args, **kwds)


def log(level, msg, *args, **kwds):
    return logging.getLogger('sim').log(level, msg, *args, **kwds)


def exception(msg, *args, **kwds):
    return logging.getLogger('sim').exception(msg, *args, **kwds)



from .log import (register_custom_log, CustomLogger, LogFmtFilter, conf_log,
                  core_log, gear_log, set_log_level, typing_log, util_log)
from .registry import (PluginBase, clear, reset, load_plugin_folder, inject,
                       Inject, MayInject, inject_async, reg)
from .trace import pygears_excepthook, register_issue, MultiAlternativeError
from .trace_format import TraceLevel, enum_traceback

__all__ = [
    'PluginBase', 'reg', 'clear', 'load_plugin_folder', 'core_log',
    'typing_log', 'util_log', 'gear_log', 'conf_log', 'CustomLogger',
    'LogFmtFilter', 'pygears_excepthook', 'TraceLevel', 'enum_traceback',
    'set_log_level', 'inject', 'Inject', 'MayInject', 'MultiAlternativeError',
    'register_custom_log'
]



import ast
from . import Context, FuncContext, node_visitor, ir, visit_ast, GearContext
from pygears.core.gear import InSig, OutSig
from pygears.typing import is_type
from .cast import resolve_cast_func


def extend_stmts(stmts, extension):
    if extension:
        if not isinstance(extension, list):
            if isinstance(extension, ir.Statement):
                stmts.append(extension)
        else:
            for e in extension:
                if isinstance(e, ir.Statement):
                    stmts.append(e)


def output_port_shadow_check(name, ctx):
    if (isinstance(ctx, GearContext) and name in ctx.intfs and ctx.intfs[name].dtype.direction):
        raise SyntaxError(f'Variable "{name}" has the same name as the output'
                          f'interface. This is currently not supported')


def infer_targets(ctx, target, source):
    dtype = source.dtype
    if isinstance(target, ir.Name):
        ctx.alias_map[target.name] = source
        if target.name not in ctx.scope:
            var = ir.Variable(target.name, dtype, reg=target.name in ctx.registers)
            ctx.scope[target.name] = var
            target.obj = var
        else:
            output_port_shadow_check(target.name, ctx)

    elif isinstance(target, ir.ConcatExpr):
        # We can only make this check if the value is recognized PyGears type.
        # If it is some random Python type, just hope for the best
        if is_type(dtype) and len(dtype) != len(target.operands):
            raise SyntaxError(
                f'Cannot unpack value of type "{dtype!r}" with {len(dtype)} component(s) into {len(target.operands)} variables: '
                f'"{target}".')

        for i, t in enumerate(target.operands):
            infer_targets(ctx, t, ir.SubscriptExpr(source, ir.ResExpr(i)))
    elif isinstance(target, ir.SubscriptExpr):
        # # TODO: can we do some check here?
        if isinstance(target.val, ir.Name):
            ctx.alias_map[target.val.name] = ctx.ref(target.val.name)
    else:
        breakpoint()


def assign_targets(ctx: Context, target, source):
    infer_targets(ctx, target, source)

    if source.dtype in (OutSig, InSig):
        ctx.scope[target.name].val = source.val
        return None

    return ir.AssignValue(target, source)


@node_visitor(ast.AnnAssign)
def _(node, ctx: Context):
    targets = visit_ast(node.target, ctx)
    annotation = visit_ast(node.annotation, ctx)

    if not (isinstance(annotation, ir.ResExpr) or isinstance(annotation.val, type)
            or is_type(annotation.val)):
        raise SyntaxError(f'Variable annotation has to be a type, not "{annotation}"')

    if node.value is None:
        output_port_shadow_check(targets.name, ctx)
        ctx.scope[targets.name] = ir.Variable(targets.name, annotation.val)
        return

    if node.value:
        init = visit_ast(node.value, ctx)

        init_cast = ir.CastExpr(init, annotation.val)
        stmts = assign_targets(ctx, targets, init_cast)
        if not isinstance(stmts, list):
            stmts = [stmts]

        # for s in stmts:
        #     s.target.obj.val = s.val
        #     if init.val is None or getattr(init.val, 'unknown', False):
        #         s.target.obj.any_init = True

        return stmts


@node_visitor(ast.AugAssign)
def _(node, ctx: Context):
    target = visit_ast(node.target, ctx)
    value = visit_ast(node.value, ctx)

    # TODO: We should probably invoke truncate function to preserve the sign correctly
    return ir.AssignValue(
        target, ir.CastExpr(ir.BinOpExpr((ctx.ref(target.name), value), type(node.op)),
                            target.dtype))


@node_visitor(ast.Assign)
def _(node, ctx: Context):
    value = visit_ast(node.value, ctx)

    stmts = []
    for t in node.targets:
        targets = visit_ast(t, ctx)
        extend_stmts(stmts, assign_targets(ctx, targets, value))

    return stmts


@node_visitor(ast.Assert)
def _(node, ctx: Context):
    test = visit_ast(node.test, ctx)
    msg = node.msg.s if node.msg else 'Assertion failed.'
    return ir.Assert(test, msg=msg)


@node_visitor(ast.Break)
def _(node, ctx: Context):
    return ir.Jump('break')


@node_visitor(ast.Return)
def _(node: ast.Return, ctx: FuncContext):
    expr = visit_ast(node.value, ctx)

    if not isinstance(ctx, FuncContext):
        raise Exception('Return found outside function')

    # TODO: If there are multiple returns from a function, return types might
    # clash which is not supported right now. Check that this is not the case
    if ctx.ret_dtype is not None:
        expr = resolve_cast_func(expr, ctx.ret_dtype)
    else:
        ctx.ret_dtype = expr.dtype

    return ir.FuncReturn(ctx.funcref, expr)


@node_visitor(ast.Pass)
def _(node: ast.Pass, ctx: Context):
    return []


@node_visitor(ast.ImportFrom)
def _(node: ast.ImportFrom, ctx: Context):
    # TODO: Handle this properly
    return []



import ast
import sys
import inspect
import math
from functools import partial
from . import Context, FuncContext, ir, node_visitor, visit_ast
from .inline import form_gear_args, call_gear, parse_func_call
from .cast import resolve_cast_func
from pygears import reg

from pygears.core.partial import Partial, MultiAlternativeError

from pygears.typing import code, div
from pygears.typing import is_type, typeof, Tuple, Array
from pygears.typing import cast, floor
from pygears.typing.queue import QueueMeta

from pygears.core.gear_inst import gear_signature, infer_params, get_function_context_dict, TypeMatchError, TooManyArguments, GearArgsNotSpecified


def parse_func_args(args, kwds, ctx):
    if args is None:
        args = []

    if kwds is None:
        kwds = []

    func_args = []
    for arg in args:
        if isinstance(arg, ast.Starred):
            var = visit_ast(arg.value, ctx)

            if isinstance(var, ir.ResExpr) and isinstance(var.val, list):
                func_args.extend(var.val)
            else:
                for i in range(len(var.dtype)):
                    func_args.append(ir.SubscriptExpr(val=var, index=ir.ResExpr(i)))

        else:
            func_args.append(visit_ast(arg, ctx))

    func_kwds = {kwd.arg: visit_ast(kwd.value, ctx) for kwd in kwds}

    return func_args, func_kwds


def get_gear_signatures(func, args, kwds):
    alternatives = [func] + getattr(func, 'alternatives', [])

    signatures = []
    kwds = {n: v.val if isinstance(v, ir.ResExpr) else v for n, v in kwds.items()}

    for f in alternatives:
        try:
            args_dict, templates = gear_signature(f, args, kwds)
        except (TooManyArguments, GearArgsNotSpecified):
            pass
        else:
            signatures.append((f, args_dict, templates))

    return signatures


def const_func_args(args, kwds):
    return (all(isinstance(node, ir.ResExpr) for node in args)
            and all(isinstance(node, ir.ResExpr) for node in kwds.values()))


def resolve_compile_time(func, args, kwds):
    # If all arguments are resolved expressions, maybe we can evaluate the
    # function at compile time
    return ir.ResExpr(func(*(a.val for a in args), **{n: v.val for n, v in kwds.items()}))


def resolve_gear_alternative(func, args, kwds):
    errors = []
    for f, args, templates in get_gear_signatures(func, args, kwds):
        try:
            params = infer_params(args, templates, get_function_context_dict(f))
        except TypeMatchError:
            errors.append((f, *sys.exc_info()))
        else:
            return (f, params)

    raise MultiAlternativeError(errors)


def resolve_gear_call(func, args, kwds):
    args, kwds = form_gear_args(args, kwds, func)

    f, params = resolve_gear_alternative(func, args, kwds)

    return ir.CallExpr(f, args, kwds, params)


compile_time_builtins = {
    all, max, int, len, type, isinstance, div, floor, cast, QueueMeta.sub, Array.code, Tuple.code,
    code, is_type, typeof
}


class Super:
    def __init__(self, cls, val):
        self.cls = cls
        self.val = val

    def __getattribute__(self, name):
        cls = super().__getattribute__('cls')
        val = super().__getattribute__('val')
        for base in inspect.getmro(cls):
            if name in base.__dict__:
                if base is object:
                    return getattr(base, name)

                f = base.__dict__[name]
                f = partial(f, val)

                return f
        else:
            raise AttributeError(name)


def get_class_that_defined_method(meth):
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing

    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        if isinstance(cls, type):
            return cls

    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


special_funcs = {
    abs: '__abs__',
    math.ceil: '__ceil__',
    math.floor: '__floor__',
    round: '__round__',
}


def hashable(v):
    """Determine whether `v` can be hashed."""
    try:
        hash(v)
    except TypeError:
        return False
    return True


def resolve_func(func, args, kwds, ctx):
    if is_type(func):
        if const_func_args(args, kwds):
            return resolve_compile_time(func, args, kwds)

        return resolve_cast_func(args[0], func)

    hashable_func = hashable(func)

    if isinstance(func, partial):
        args = func.args + tuple(args)
        func = func.func
    elif not inspect.isbuiltin(func) and hasattr(func, '__self__'):
        if func is super:
            if not isinstance(ctx, FuncContext):
                raise Exception(f'super() called outside a method function')

            f = ctx.funcref.func
            obj = ctx.ref('self')
            cls = get_class_that_defined_method(f).__base__

            return ir.ResExpr(Super(cls, obj))
        elif getattr(func, '__func__', None) in reg['hls/ir_builtins']:
            val = func.__self__
            for name, v in ctx.scope.items():
                if isinstance(v, ir.Variable) and val is v.val:
                    val = ctx.ref(name)
                    break

            return reg['hls/ir_builtins'][func.__func__](val, *args, **kwds)
        elif const_func_args(args, kwds):
            return resolve_compile_time(func, args, kwds)
        elif is_type(func.__self__):
            if func.__name__ == 'decode':
                return ir.CastExpr(args[0], ir.ResExpr(func.__self__))
            else:
                breakpoint()
                raise Exception
        else:
            breakpoint()
            func = func.__func__
            # raise Exception

    if hashable_func and func not in reg['hls/ir_builtins']:
        if func in special_funcs:
            return resolve_func(getattr(args[0].dtype, special_funcs[func]), args, kwds, ctx)

    if isinstance(func, Partial):
        intf, stmts = call_gear(func, *form_gear_args(args, kwds, func), ctx)
        ctx.ir_parent_block.stmts.extend(stmts)
        return intf

    if hashable_func and func in compile_time_builtins and const_func_args(args, kwds):
        return resolve_compile_time(func, args, kwds)

    if hashable_func and func in reg['hls/ir_builtins']:
        try:
            return reg['hls/ir_builtins'][func](*args, **kwds)
        except TypeError as e:
            breakpoint()
            raise SyntaxError(str(e).replace('<lambda>()', repr(func)))

    if const_func_args(args, kwds):
        return resolve_compile_time(func, args, kwds)
    elif hasattr(func, 'dispatch'):
        return resolve_func(func.dispatch(args[0].dtype), args, kwds, ctx)
    else:
        return parse_func_call(func, args, kwds, ctx)


@node_visitor(ast.Call)
def _(node, ctx: Context):
    name = visit_ast(node.func, ctx)

    assert isinstance(name, (ir.ResExpr, ir.AttrExpr))

    args, kwds = parse_func_args(node.args, node.keywords, ctx)

    if not isinstance(name, ir.ResExpr):
        return ir.CallExpr(name, args, kwds)

    func = name.val

    return resolve_func(func, args, kwds, ctx)



def remove_node(node):
    node.prev[0].next = node.next[:]
    node.next[0].prev = node.prev[:]


def insert_node_before(origin, node):
    i = origin.prev[0].next.index(origin)
    origin.prev[0].next[i] = node

    node.prev = origin.prev[:]

    node.next = [origin]
    origin.prev = [node]

    return node


def insert_node_after(origin, node):
    node.next = origin.next[:]
    node.next[0].prev = [node]

    node.prev = [origin]
    origin.next = [node]

    return node



import fnmatch
import functools
import hashlib

from pygears import reg
from .base_resolver import ResolverTypeError
from . import hdl_log
from pygears.hdl import mod_lang, hdlmod
from pygears.util.fileio import save_file


def path_name(path):
    if path.startswith('/'):
        path = path[1:]

    full_name = path.replace('/', '_')
    if len(full_name) > 100:
        path_l = path.split('/')
        head = '_'.join(path_l[:3])
        tail = '_'.join(path_l[-3:])
        mid = '_'.join(path_l[3:-3])
        full_name = head + '_' + hashlib.sha1(mid.encode()).hexdigest()[:8] + '_' + tail

    return full_name


class HDLModuleInst:
    def __init__(self, node, lang=None, resolver=None):
        self.node = node

        if lang is None:
            lang = mod_lang(node)

        self.lang = lang

        self._impl_parse = None
        self.memoized = False

        if 'memoized' in self.node.params:
            memnode = self.node.params['memoized']

            # TODO: What if hdlmod is a part of different cosim build and
            # located at different folder. We should include that folder in the
            # path

            # TODO: What if hdlmod hasn't been
            # generated? This can happen if we only generate a part of the
            # design
            if memnode in reg['hdlgen/map']:
                self.memoized = True
                hdlmod = reg['hdlgen/map'][memnode]
                self.resolver = hdlmod.resolver
                return

        if resolver is not None:
            self.resolver = resolver
        else:
            self.resolver = self.get_resolver()

    def get_resolver(self):
        if self.node.parent is None:
            return reg[f'{self.lang}gen/dflt_resolver'](self.node)

        for r in reg[f'{self.lang}gen/resolvers']:
            try:
                return r(self.node)
            except ResolverTypeError:
                pass
        else:
            resolver = reg[f'{self.lang}gen/dflt_resolver'](self.node)
            hdl_log().warning(
                f'Unable to compile "{self.node.name}" to HDL and no HDL module with the name '
                f'"{resolver.module_name}" found on the path. Module connected as a black-box.')
            return resolver

    @property
    def _basename(self):
        return self.basename

    @property
    def basename(self):
        return self.node.basename

    @property
    @functools.lru_cache()
    def traced(self):
        def check(pattern):
            if isinstance(pattern, str):
                return fnmatch.fnmatch(self.node.name, pattern)
            else:
                return pattern(self.node)

        self_traced = any(check(p) for p in reg['debug/trace'])

        if self.hierarchical:
            children_traced = any(hdlmod(child).traced for child in self.node.child)
        else:
            children_traced = False

        return self_traced or children_traced

    @property
    def hierarchical(self):
        return self.node.meta_kwds.get('hdl', {}).get('hierarchical', self.node.hierarchical)

    @property
    def hier_path_name(self):
        return path_name(self.node.name)

    @property
    def inst_name(self):
        return path_name(self.node.basename)

    @property
    def module_name(self):
        return self.resolver.module_name

    @property
    def wrap_module_name(self):
        module_name = self.module_name
        if self.wrapped:
            return f'{module_name}_{self.parent_lang}_wrap'

        return module_name

    @property
    def wrap_file_name(self):
        return f'{self.wrap_module_name}.{self.parent_lang}'

    @property
    def file_basename(self):
        return self.resolver.file_basename

    @property
    def files(self):
        res_files = self.resolver.files

        if self.wrapped:
            res_files.append(self.wrap_file_name)

        return res_files

    @property
    def parent_lang(self):
        return mod_lang(self.node.parent)

    @property
    def wrapped(self):
        if mod_lang(self.node.parent) != self.lang:
            return True

        if self.node is reg['hdl/top'] and self.params:
            return True

        return False

    @property
    def params(self):
        return self.resolver.params

    def generate(self, template_env, outdir):
        if not self.memoized:
            self.resolver.generate(template_env, outdir)

        if not self.node.parent:
            return

        # TODO: What about reusing memoized module that didn't need a
        # wrapper. Discern this.
        if self.wrapped:
            save_file(self.wrap_file_name, outdir, self.get_wrap(self.parent_lang))



from pygears import alternative, gear
from pygears.lib import cart
from pygears.typing import Queue, Uint, Tuple, Bool


@gear
async def clip(din: Queue[Tuple[{
        'data': 't_data',
        'size': Uint
}]], *, init=1) -> Queue['t_data']:
    """Clips the input transaction into two separate transactions by
    sending the ``eot`` after a given number of data has passed (specified by
    the ``size`` field of the :class:`Tuple`)

    Args:
        init: Initialization value for the counter

    Returns:
        A :class:`Queue` type whose data consists of the ``data`` field of
          the :class:`Tuple` input
    """
    cnt = din.dtype.data['size'](init)
    pass_eot = Bool(True)

    async for ((data, size), eot) in din:
        yield (data, eot or ((cnt == size) and pass_eot))

        if ((cnt == size) and pass_eot):
            # to prevent wraparound if counter overflows
            pass_eot = 0
        cnt += 1


@alternative(clip)
@gear
def clip2(din: Queue, size: Uint, *, init=1) -> b'din':
    return cart(size, din, order=[1, 0]) | clip(init=init)
